"""Message printing functions. cddm uses cddm.conf.CDDMConfig.verbose 
variable to define verbosity level and behavior of printing functions.

verbosity levels 1 and 2 enable printing, 0 disables printing messages.
 """

from __future__ import absolute_import, print_function, division
import time
import dtmm.conf

def print_message(message, level = 1):
    import warnings
    warnings.warn("Deprecated! use print1 or print2")
    if level >= 1:
        print(message)
    
def print_layer_rate(n_frames, t0, t1 = None, message = "... processed", level = None):
    if level is not None:
        import warnings
        warnings.warn("Deprecated use of level argument", DeprecationWarning)
    if dtmm.conf.DTMMConfig.verbose >= 2:
        if t1 is None:
            t1 = time.time()#take current time
        print (message + " {0} layers with an average rate {1:.2f}".format(n_frames, n_frames/(t1-t0)))
        
def print1(*args, **kwargs):
    """prints level 1 messages"""
    if dtmm.conf.DTMMConfig.verbose >= 1:
        print(*args,**kwargs)
        
def print2(*args,**kwargs):
    """prints level 2 messages"""
    if dtmm.conf.DTMMConfig.verbose >= 2:
        print(*args,**kwargs)
        
def print_progress (iteration, total, prefix = '', suffix = '', decimals = 1, length = 50, fill = '=', level = None):
    """
    Call in a loop to create terminal progress bar
    """
    if level is not None:
        import warnings
        warnings.warn("Deprecated use of level argument", DeprecationWarning)
    if dtmm.conf.DTMMConfig.verbose >= 1:
        percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + '-' * (length - filledLength)
        print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end = '\r')
    #        sys.stdout.write(s)
    #        sys.stdout.flush()
        # Print New Line on Complete
        if iteration == total: 
            print()
        
def print_frame_rate(n_frames, t0, t1 = None, message = "... processed"):
    """Prints calculated frame rate"""
    if dtmm.conf.CDDMConfig.verbose >= 2:
        if t1 is None:
            t1 = time.time()#take current time
        print (message + " {0} frames with an average frame rate {1:.2f}".format(n_frames, n_frames/(t1-t0)))
        
def disable_prints():
    """Disable message printing. Returns previous verbosity level"""
    #set verbosity to negative value
    value = dtmm.conf.CDDMConfig.verbose
    dtmm.conf.CDDMConfig.verbose = - abs(value)
    return value
    
def enable_prints(level = None):
    """Enable message printing. Returns previous verbosity level"""
    #set verbosity to positve value
    value = dtmm.conf.CDDMConfig.verbose if level is None else int(level)
    dtmm.conf.CDDMConfig.verbose =  abs(dtmm.conf.CDDMConfig.verbose)
    return value

def test_progress():
    import time
    print("start")
    for i in range(100):
        print_progress(i,100, fill = "=")
        time.sleep(0.02)
    print_progress(100,100)
    print("stop")


# -*- coding: utf-8 -*-
"""
Wave creation and wave characterization functions.
"""

from __future__ import absolute_import, print_function, division

import numpy as np
from dtmm.conf import FDTYPE,CDTYPE, cached_function, BETAMAX
import dtmm.fft as fft

def betax1(n, k0, out = None):
    """Returns beta values of all possible plane eigenwaves in 2D.
    
    Parameters
    ----------
    n : int
        Size of the planewave
    k0 : float
        Wavenumber in pixel units.
    out : ndarray, optional
        Output arrays tuple
    
    Returns
    -------
    array
        beta array  
    """
    k0 = np.abs(np.asarray(k0, FDTYPE)[...,np.newaxis]) #make it broadcastable
    xx = np.asarray(np.fft.fftfreq(n), dtype = FDTYPE)
    beta = np.multiply((2 * np.pi / k0) , xx, out = out)
    return beta

def betaxy2betaphi(betax, betay):
    phi  = np.arctan2(betay, betax)
    beta = np.sqrt(betax**2 + betay**2)
    return beta, phi

def betaxy2beta(betax, betay):
    return np.sqrt(betax**2 + betay**2)

def betaxy2phi(betax, betay):
    return np.arctan2(betay, betax)

def betaphi2betaxy(beta, phi):
    betax = beta*np.cos(phi)
    betay = beta*np.sin(phi)
    return betax, betay

def betaphi2betax(beta, phi):
    return beta*np.cos(phi)

def betaphi2betay(beta, phi):
    return beta*np.sin(phi)


def betaphi(shape, k0, out = None):
    """Returns beta and phi arrays of all possible plane eigenwaves.
    
    Parameters
    ----------
    shape : (int,int)
        Shape of the plane eigenwave.
    k0 : float
        Wavenumber in pixel units.
    out : (ndarray, ndarray), optional
        Output arrays tuple
    
    Returns
    -------
    array, array
        beta, phi arrays    
    """
    if out is None:
        out = None, None
    k0 = np.abs(np.asarray(k0, FDTYPE)[...,np.newaxis,np.newaxis]) #make it broadcastable
    ay, ax = map(lambda x : np.asarray(np.fft.fftfreq(x), dtype = FDTYPE), shape[-2:])
    xx, yy = np.meshgrid(ax, ay,copy = False, indexing = "xy") 
    beta = np.multiply((2 * np.pi / k0) , np.sqrt(xx**2 + yy**2), out = out[0])
    phi  = np.arctan2(yy,xx, out = out[1])
    return beta, phi

@cached_function
def betaxy(shape, k0, out = None):
    """Returns betax, betay arrays of plane eigenwaves.
    
    Parameters
    ----------
    shape : (int,int)
        Shape of the plane eigenwave.
    k0 : float
        Wavenumber in pixel units.
    out : (ndarray, ndarray), optional
        Output arrays tuple
    
    Returns
    -------
    array, array
        beta, phi arrays      
    """
    #ax, ay = map(np.fft.fftfreq, shape,(d,)*len(shape))
    k0 = np.asarray(k0,dtype = FDTYPE)[...,np.newaxis,np.newaxis] #make it broadcastable
    ay, ax = map(lambda x : np.asarray(np.fft.fftfreq(x), dtype = FDTYPE), shape[-2:])
    xx, yy = np.meshgrid(ax, ay,copy = False, indexing = "xy") 
    if out is None:
        out = None, None
    l = (2 * np.pi / k0)
    return np.multiply(l,xx, out = out[0]), np.multiply(l,yy, out = out[1])

def k0(wavelength, pixelsize = 1.):
    """Calculate wave number in vacuum from a given wavelength and pixelsize
    
    Parameters
    ----------
    wavelength : float or array of floats
        Wavelength in nm
    pixelsize: float
        Pixelsize in nm
    
    Returns
    -------
    array
        Wavenumber array     
    """
    out = 2*np.pi/np.asarray(wavelength) * pixelsize
    return np.asarray(out, dtype = FDTYPE)

def wavelengths(start = 380,stop = 780, count = 9):
    """Raturns wavelengths (in nanometers) equaly spaced in wavenumbers between 
    start and stop.
    
    Parameters
    ----------
    start : float
        First wavelength
    stop : float
        Last wavelength    
    count : int
        How many wavelengths
        
    Returns
    -------
    out : ndarray
        A wavelength array
    """
    out = 1./np.linspace(1./start, 1./stop, count)
    return np.asarray(out, dtype = FDTYPE)
    

def eigenwave(shape, i, j, amplitude = None, out = None):
    """Returns a planewave with a given fourier coefficient indices i and j. 
    
    Parameters
    ----------
    shape : (int,int)
        Shape of the plane eigenwave.
    i : int
        i-th index of the fourier coefficient 
    j : float
        j-th index of the fourier coefficient 
    amplitude : complex
        Amplitude of the fourier mode.
    out : ndarray, optional
        Output array
    
    Returns
    -------
    array
        Plane wave array.       
    """    
    if out is None:
        f = np.zeros(shape, dtype = CDTYPE)
    else:
        f = np.asarray(out)
        f[...] = 0.
    if amplitude is None:
        amplitude = np.multiply.reduce(shape[-2:])
    f[...,i,j] = amplitude
    return fft.ifft2(f, out = out)

def eigenwave1(n, i,  amplitude = None):
    """Returns a planewave with a given fourier coefficient indices i 
    
    Parameters
    ----------
    n : int
        Shape of the plane eigenwave.
    i : int
        i-th index of the fourier coefficient 
    amplitude : complex
        Amplitude of the fourier mode.
    out : ndarray, optional
        Output array
    
    Returns
    -------
    array
        Plane wave array.       
    """    
    f = np.zeros((n,), dtype = CDTYPE)
    if amplitude is None:
        amplitude = n
    f[...,i] = amplitude
    return fft.ifft(f,overwrite_x = True)


@cached_function
def eigenmask(shape, k0, betamax = BETAMAX):
    b,p = betaphi(shape, k0)
    mask = b <= betamax
    return mask

@cached_function
def eigenmask1(n, k0, betamax = BETAMAX):
    b = betax1(n, k0)
    mask = np.abs(b) <= betamax
    return mask

def mask2beta(mask,k0):
    shape = mask.shape[-2:]
    b,p = betaphi(shape, k0)
    if mask.ndim == 3:
        return tuple([b[i][mask[i]] for i in range(mask.shape[0])])
    else:
        return b[mask]
    
def mask2beta1(mask,k0):
    n = mask.shape[-1]
    b = betax1(n, k0)
    if mask.ndim == 2:
        return tuple([b[i][mask[i]] for i in range(mask.shape[0])])
    else:
        return b[mask]
    
def mask2phi(mask,k0):
    shape = mask.shape[-2:]
    b,p = betaphi(shape, k0)
    if mask.ndim == 3:
        return tuple([p[mask[i]] for i in range(mask.shape[0])])
    else:
        return p[mask]

def mask2indices(mask, k0):
    shape = mask.shape[-2:]
    ii, jj = np.meshgrid(range(shape[-2]), range(shape[-1]),copy = False, indexing = "ij") 
    if mask.ndim == 3: #multiwavelength
        out = (_get_indices_array(ii,jj, mask[i]) for i in range(mask.shape[0]))
        return tuple(out)
    else:
        return _get_indices_array(ii,jj, mask)  

def mask2indices1(mask, k0):
    n = mask.shape[-1]
    ii = np.arange(n)
    if mask.ndim == 2: #multiwavelength
        out = (ii[mask[i]] for i in range(mask.shape[0]))
        return tuple(out)
    else:
        return ii[mask]

@cached_function
def eigenbeta(shape, k0, betamax = BETAMAX):
    b,p = betaphi(shape, k0)
    mask = b <= betamax
    if mask.ndim == 3:
        bp = tuple([b[i][mask[i]] for i in range(mask.shape[0])])
        return bp
    else:
        b = b[mask]
        return b
    
@cached_function
def eigenbeta1(n, k0, betamax = BETAMAX):
    b = betax1(n, k0)
    mask = np.abs(b) <= betamax
    if mask.ndim == 2:
        return tuple([b[i][mask[i]] for i in range(mask.shape[0])])
    else:
        return b[mask]
    
@cached_function
def eigenphi(shape, k0, betamax = BETAMAX):
    b,p = betaphi(shape, k0)
    mask = b <= betamax
    if mask.ndim == 3:
        pp = tuple([p[mask[i]] for i in range(mask.shape[0])])
        return pp
    else:
        p = p[mask]
        return p

def _get_indices_array(ii,jj, mask):
    itmp = ii[mask]
    jtmp = jj[mask]
    out = np.empty(shape = itmp.shape + (2,), dtype = itmp.dtype)
    out[:,0] = itmp
    out[:,1] = jtmp
    return out

@cached_function
def eigenindices(shape, k0, betamax = BETAMAX):
    ii, jj = np.meshgrid(range(shape[-2]), range(shape[-1]),copy = False, indexing = "ij") 
    mask = eigenmask(shape, k0, betamax)
    if mask.ndim == 3: #multiwavelength
        out = (_get_indices_array(ii,jj, mask[i]) for i in range(mask.shape[0]))
        return tuple(out)
    else:
        return _get_indices_array(ii,jj, mask)
 
@cached_function
def eigenindices1(n, k0, betamax = BETAMAX):
    ii = np.arange(n) 
    mask = eigenmask1(n, k0, betamax)
    if mask.ndim == 2: #multiwavelength
        out = (ii[mask[i]] for i in range(mask.shape[0]))
        return tuple(out)
    else:
        return ii[mask]
    
#@cached_function
#def eigenwaves(shape, k0, betamax = BETAMAX):
#    ii, jj = eigenindices(shape, k0, betamax)
#    out = np.empty(shape = (len(ii),) + shape, dtype = CDTYPE)
#    for n in range(len(ii)):
#        i = ii[n]
#        j = jj[n]
#        eigenwave(shape, i, j,  amplitude = 1, out =  out[n])
#    return out

def wave2eigenwave(wave, out = None):
    """Converts any wave to nearest eigenwave"""
    wave = np.asarray(wave)
    if out is None:
        out = np.empty(shape = wave.shape, dtype = CDTYPE)
    
    shape = wave.shape
    assert wave.ndim >= 2
    
    if len(shape) > 2:
        h = shape[-2]
        w = shape[-1]
        wave = wave.reshape(-1) #flatten
        n = len(wave)//h//w
        wave = wave.reshape((n,h,w))
        o = out.reshape((n,h,w))
    else:
        o = out
    
    s = np.abs(fft.fft2(wave))**2
    
    if len(shape) == 2:
        s = s[None,...]
        o = out[None,...]

    for i,a in enumerate(s):
        avg = a.sum()
        indices = np.argmax(a)
        ii,jj = np.unravel_index(indices,a.shape)
        eigenwave(a.shape,ii,jj, amplitude = avg**0.5, out = o[i])
    
    return out
    

def planewave(shape, k0, beta , phi, out = None):
    """Returns a 2D planewave array with a given beta, phi, wave number k0.
    
    Parameters
    ----------
    shape : (int,int)
        Shape of the plane eigenwave.
    k0 : float or array of floats
        Wavenumbers in pixel units.
    beta : float
        Beta parameter of the plane wave
    phi: float
        Phi parameter of the plane wave
    out : (ndarray, ndarray), optional
        Output arrays tuple
    
    Returns
    -------
    array
        Plane wave array.       
    """
    k0 = np.asarray(k0)[...,np.newaxis,np.newaxis] #make it broadcastable
    beta = np.asarray(beta)[...,np.newaxis,np.newaxis]
    phi = np.asarray(phi)[...,np.newaxis,np.newaxis]
    ay, ax = [np.arange(-l // 2 + 1., l // 2 + 1.) for l in shape[-2:]]
    xx, yy = np.meshgrid(ax, ay, indexing = "xy", copy = False)
    xx = np.asarray(xx, dtype = FDTYPE)
    yy = np.asarray(yy, dtype = FDTYPE)
    kx = np.asarray(k0*beta*np.cos(phi), dtype = FDTYPE)
    ky = np.asarray(k0*beta*np.sin(phi), dtype = FDTYPE)
    out = np.exp((1j*(kx*xx+ky*yy)), out = out)
    return np.divide(out,out[...,0,0][...,None,None],out)

def planewave1(n, k0, beta, out = None):
    """Returns a 1D planewave array with a given beta, wave number k0.
    
    Parameters
    ----------
    n : int
        Shape of the plane eigenwave.
    k0 : float or array of floats
        Wavenumbers in pixel units.
    beta : float
        Beta parameter of the plane wave
    
    Returns
    -------
    array
        Plane wave array.       
    """
    k0 = np.asarray(k0)[...,np.newaxis] #make it broadcastable
    beta = np.asarray(beta)[...,np.newaxis]
    xx = np.arange(-n // 2 + 1., n // 2 + 1.)
    xx = np.asarray(xx, dtype = FDTYPE)
    kx = np.asarray(k0*beta, dtype = FDTYPE)
    out = np.exp((1j*(kx*xx)), out = out)
    return np.divide(out,out[...,0][...,None],out)

__all__ = [ "betaphi","betaxy","eigenwave","planewave","k0","wavelengths"]


"""Tests for fft"""

import unittest
import numpy as np
from dtmm.fft import fft2, ifft2
import dtmm.conf
from dtmm.conf import MKL_FFT_INSTALLED

class TestFFT(unittest.TestCase):
    
    def setUp(self):
        a0 = np.random.randn(46,64)+0j
        af0 = np.fft.fft2(a0)
        afi0 = np.fft.ifft2(a0)        
        
        a = np.random.randn(2,2,46,64)+0j
        af = np.fft.fft2(a)
        afi = np.fft.ifft2(a)
        
        self._farrays = [(a0,af0),(a,af)]
        self._fiarrays = [(a0,afi0),(a,afi)]
     
    def _assert_fft_inplace(self, fftfunc, inarray, result):
        a = inarray.copy()
        out = fftfunc(a,a)
        self.assertTrue(np.allclose(out, result))

    def _assert_fft_out(self, fftfunc, inarray, result):
        a = inarray.copy()
        out = np.empty_like(a)
        out = fftfunc(a,out)
        self.assertTrue(np.allclose(out, result)) 

    def _assert_fft_new(self, fftfunc, inarray, result):
        a = inarray.copy()
        out = fftfunc(a)
        self.assertTrue(np.allclose(out, result))  
        
    def _assert_fft(self, fftfunc, inarray, result):
        self._assert_fft_new(fftfunc, inarray,result)
        self._assert_fft_out(fftfunc, inarray,result)
        self._assert_fft_inplace(fftfunc, inarray,result)
        
    def test_mkl_fft2(self):
        if MKL_FFT_INSTALLED:
            dtmm.conf.set_fftlib("mkl_fft")
            for a,out in self._farrays:
                self._assert_fft(fft2, a, out)

    def test_mkl_ifft2(self):
        if MKL_FFT_INSTALLED:
            dtmm.conf.set_fftlib("mkl_fft")
            for a,out in self._fiarrays:
                self._assert_fft(ifft2, a, out)
        
    def test_np_fft2(self):
        dtmm.conf.set_fftlib("numpy")
        for a,out in self._farrays:
            self._assert_fft(fft2, a, out)

    def test_np_ifft2(self):
        dtmm.conf.set_fftlib("numpy")
        for a,out in self._fiarrays:
            self._assert_fft(ifft2, a, out)
    
    def test_sp_fft2(self):
        dtmm.conf.set_fftlib("scipy")
        for a,out in self._farrays:
            self._assert_fft(fft2, a, out)

    def test_sp_ifft2(self):
        dtmm.conf.set_fftlib("scipy")
        for a,out in self._fiarrays:
            self._assert_fft(ifft2, a, out)

                
if __name__ == "__main__":
    unittest.main()


"""
4x4 transfer matrix method functions for 2d data
"""

from __future__ import absolute_import, print_function, division
import numpy as np

from dtmm.conf import  BETAMAX, CDTYPE, DTMMConfig
from dtmm.linalg import dotmdm, inv, dotmv, bdotmm, bdotmd, bdotdm
from dtmm.print_tools import print_progress

import dtmm.tmm as tmm
from dtmm.tmm import alphaffi, phase_mat

from dtmm.wave import eigenbeta1, eigenindices1, eigenmask1, eigenwave1, betaxy2beta, mask2beta1, mask2indices1,betaxy2phi
from dtmm.wave import k0 as wavenumber
from dtmm.field import field2modes1, modes2field1
from dtmm.fft import mfft

def layer_mat2d(k0, d, epsv,epsa, betay = 0., method = "4x4", mask = None):
    """Computes characteristic matrix of a single layer M=F.P.Fi,
    
    Numpy broadcasting rules apply.
    
    Parameters
    ----------
    k0 : float or sequence of floats
        A scalar or a vector of wavenumbers
    d : array_like
        Layer thickness
    epsv : array_like
        Epsilon eigenvalues.
    epsa : array_like
        Optical axes orientation angles (psi, theta, phi).
    method : str, optional
        Either a 4x4 or 4x4_1
    
    Returns
    -------
    cmat : ndarray
        Characteristic matrix of the layer.
    """
    if method not in ("4x4","4x4_1"):
        raise ValueError("Unsupported method: '{}'".format(method))
    k0 = np.asarray(k0)
    shape = epsv.shape[-2]
    if mask is None:
        mask = eigenmask1(shape, k0)
        betas = eigenbeta1(shape, k0)
        indices = eigenindices1(shape, k0)
    else:
        betas = mask2beta1(mask,k0)
        indices = mask2indices1(mask,k0)
    if k0.ndim == 0:
        return _layer_mat2d(k0,d,epsv,epsa, mask, betas,betay, indices, method)
    else:
        out = (_layer_mat2d(k0[i],d,epsv,epsa, mask[i],betas[i], betay, indices[i],method) for i in range(len(k0)))
        return tuple(out)


def _layer_mat2d(k0,d,epsv,epsa, mask, betaxs, betay,indices, method):   
    n = len(betaxs)
    kd = k0*d
    shape = epsv.shape[-2]
    out = np.empty(shape = (n, n, 4, 4), dtype = CDTYPE)
    beta = betaxy2beta(betaxs,betay)
    phi = betaxy2phi(betaxs,betay)
    for j,(beta,phi) in enumerate(zip(beta, phi)):   
        alpha,f,fi = alphaffi(beta,phi,epsv,epsa)
        pmat = phase_mat(alpha,-kd)
        if method == "4x4_1":
            pmat[...,1::2] = 0.
        elif method != "4x4":
            raise ValueError("Unsupported method!")

        wave = eigenwave1(shape, indices[j], amplitude = 1.)
        
        #m is shape (...,4,4)
        m = dotmdm(f,pmat,fi) 
        
        #wave is shape (...) make it broadcastable to (...,4,4)

        mw = m*wave[...,None,None]
        mf = mfft(mw, overwrite_x = True)
        mf = mf[mask,...]
        
        out[:,j,:,:] = mf

        #for i,mfj in enumerate(mf):
        #    out[i,j,:,:] = mfj

    return out

def stack_mat2d(k,d,epsv,epsa, betay = 0., method = "4x4" ,mask = None):
    n = len(d)
    indices = range(n)
    if method.startswith("2x2"):
        indices = reversed(indices)
    verbose_level = DTMMConfig.verbose
    if verbose_level > 1:
        print ("Building stack matrix.")
    for i in range(n):
        print_progress(i,n,level = verbose_level) 
        mat = layer_mat2d(k,d[i],epsv[i],epsa[i], betay = betay, mask = mask, method = method)

        if i == 0:
            if isinstance(mat, tuple):
                out = tuple((m.copy() for m in mat))
            else:
                out = mat.copy()
        else:
            if isinstance(mat, tuple):
                out = tuple((bdotmm(o,m) for o,m in zip(out,mat)))
            else:
                out = bdotmm(out,mat)

    print_progress(n,n,level = verbose_level) 

    return out 

def f_iso2d(shape, k0, n = 1., betay = 0, betamax = BETAMAX):
    k0 = np.asarray(k0)
    betax = eigenbeta1(shape,k0, betamax)
    
    if k0.ndim == 0:
        beta = betaxy2beta(betax,betay)
        phi = betaxy2phi(betax,betay)
        fmat = tmm.f_iso(n = n, beta = beta, phi = phi)
        return fmat
    else:
        beta = tuple((betaxy2beta(beta,betay) for beta in betax))
        phi = tuple((betaxy2phi(beta,betay) for beta in betax))
        fmat = (tmm.f_iso(n = n, beta = beta[i], phi = phi[i]) for i in range(len(k0)))
        return tuple(fmat)
    
def fi_iso2d(shape, k0, n = 1., betay = 0, betamax = BETAMAX):
    fmat = f_iso2d(shape, k0, n, betay, betamax)
    if isinstance(fmat, tuple):
        out = (inv(f) for f in fmat)
        return tuple(out)
    else:
        return inv(fmat)

def _system_mat2d(fmatin, cmat, fmatout):
    """Computes a system matrix from a characteristic matrix Fin-1.C.Fout"""
    fmatini = inv(fmatin)
    out = bdotdm(fmatini,cmat)
    return bdotmd(out,fmatout)  

def system_mat2d(fmatin, cmat, fmatout):
    """Computes a system matrix from a characteristic matrix Fin-1.C.Fout"""
    if isinstance(fmatin, tuple):
        if cmat is not None:
            out = (_system_mat2d(fi, c, fo) for fi,c,fo in zip(fmatin,cmat,fmatout))
        else:
            out = (_system_mat2d(fi, None, fo) for fi,fo in zip(fmatin,fmatout))
        return tuple(out)
    else:
        return _system_mat2d(fmatin, cmat, fmatout)

def _reflection_mat2d(smat):
    """Computes a 4x4 reflection matrix.
    """
    shape = smat.shape[0:-4] + (smat.shape[-4] * 4,smat.shape[-4] * 4)  
    smat = np.moveaxis(smat, -2,-3)
    smat = smat.reshape(shape)
    return tmm.reflection_mat(smat)

def reflection_mat2d(smat):
    verbose_level = DTMMConfig.verbose
    if verbose_level > 1:
        print ("Building reflectance and transmittance matrix.")    
    if isinstance(smat, tuple):
        out = []
        n = len(smat)
        for i,s in enumerate(smat):
            print_progress(i,n,level = verbose_level) 
            out.append(_reflection_mat2d(s))
        print_progress(n,n,level = verbose_level)     
        return tuple(out)
    else:
        return _reflection_mat2d(smat)

def _reflect2d(fvec_in, fmat_in, rmat, fmat_out, fvec_out = None):
    """Transmits/reflects field vector using 4x4 method.
    
    This functions takes a field vector that describes the input field and
    computes the output transmited field and also updates the input field 
    with the reflected waves.
    """
    
    fmat_ini = inv(fmat_in)
        
    avec = dotmv(fmat_ini,fvec_in)
    
    a = np.zeros(avec.shape, avec.dtype)
    a[...,0::2] = avec[...,0::2]

    if fvec_out is not None:
        fmat_outi = inv(fmat_out)
        bvec = dotmv(fmat_outi,fvec_out)
        a[...,1::2] = bvec[...,1::2] 
    else:
        bvec = np.zeros_like(avec)

    av = a.reshape(a.shape[:-2] + (a.shape[-2]*a.shape[-1],))
    out = dotmv(rmat,av).reshape(a.shape)
    
    avec[...,1::2] = out[...,1::2]
    bvec[...,::2] = out[...,::2]
        
    dotmv(fmat_in,avec,out = fvec_in)    
    return dotmv(fmat_out,bvec,out = out)

def reflect2d(fvecin, fmatin, rmat, fmatout, fvecout = None):
    """Transmits/reflects field vector using 4x4 method.
    
    This functions takes a field vector that describes the input field and
    computes the output transmited field vector and also updates the input field 
    with the reflected waves.
    """
    verbose_level = DTMMConfig.verbose
    if verbose_level > 2:
        print ("Transmitting field.")   
    if isinstance(fvecin, tuple):
        n = len(fvecin)
        if fvecout is None:
            return tuple((_reflect2d(fvecin[i], fmatin[i], rmat[i], fmatout[i]) for i in range(n)))
        else:
            return tuple((_reflect2d(fvecin[i], fmatin[i], rmat[i], fmatout[i], fvecout[i]) for i in range(n)))
    else:
        return _reflect2d(fvecin, fmatin, rmat, fmatout, fvecout)

def transfer2d(field_data_in, optical_data, betay = 0., nin = 1., nout = 1., method = "4x4", betamax = BETAMAX, field_out = None):
    
    f,w,p = field_data_in
    shape = f.shape[-1]
    d,epsv,epsa = optical_data
    k0 = wavenumber(w, p)
    
    if field_out is not None:
        mask, fmode_out = field2modes1(field_out,k0, betamax = betamax)
    else:
        fmode_out = None
    
    mask, fmode_in = field2modes1(f,k0, betamax = betamax)
    
    fmatin = f_iso2d(shape = shape, betay = betay, k0 = k0, n=nin, betamax = betamax)
    fmatout = f_iso2d(shape = shape, betay = betay, k0 = k0, n=nout, betamax = betamax)
    
    cmat = stack_mat2d(k0,d, epsv, epsa, betay = betay, mask = mask, method = method)
    smat = system_mat2d(fmatin = fmatin, cmat = cmat, fmatout = fmatout)
    rmat = reflection_mat2d(smat)
    
    fmode_out = reflect2d(fmode_in, rmat = rmat, fmatin = fmatin, fmatout = fmatout, fvecout = fmode_out)
    
    field_out = modes2field1(mask, fmode_out)
    f[...] = modes2field1(mask, fmode_in)
    
    return field_out,w,p

        
    



"""Diffractive transfer matrix method
"""

__version__ = "0.6.1"

import dtmm.conf
import numpy as np
import time
from .window import *
from .wave import *
from .linalg import * 
from .field import *
from .data import expand, rot90_director, rotate_director, cholesteric_droplet_data,load_stack, save_stack, read_raw, sphere_mask, director2data, validate_optical_data, angles2director, director2angles, read_director, refind2eps, nematic_droplet_data, nematic_droplet_director
from .color import *
from .tmm import *
from .field_viewer import field_viewer, pom_viewer
from .diffract import *
from .data_viewer import *
from .transfer import *
from .jones import jonesvec
from .rotation import rotation_matrix,rotation_matrix_x,rotation_matrix_y,rotation_matrix_z



"""
Transfer Matrix Method
======================

4x4 and 2x2 transfer matrix method functions for 1D calculation. 

The implementation is based on standard formulation of 4x4 transfer matrix method.

4x4 method
----------

Layers are stacked in the z direction, field vectors describing the field are 
f = (Ex,Hy,Ey,Hx), Core functionality is defined by field matrix calculation
functions:
    
Field vector creation/conversion functions
++++++++++++++++++++++++++++++++++++++++++

* :func:`.avec` for amplitude vector (eigenmode amplitudes).
* :func:`.fvec` for field vector creation,
* :func:`.avec2fvec` for amplitude to field conversion.
* :func:`.fvec2avec` for field to amplitude conversion.

Field matrix functions
++++++++++++++++++++++

* :func:`.f_iso` for input and output field matrix caluclation.
* :func:`.ffi_iso` computes the inverse of the field matrix. 
* :func:`.alphaf` for general field vectors and field coefficents calcualtion.
* :func:`.alphaffi` computes the inverse of the field matrix.
* :func:`.phase_mat` for phase matrix calculation.

Layer/stack computation
+++++++++++++++++++++++

* :func:`.layer_mat` for layer matrix calculation Mi=Fi.Pi.Fi^-1
* :func:`.stack_mat` for stack matrix caluclation M = M1.M2.M3....
* :func:`.system_mat` for system matrix calculation Fin^-1.M.Fout

Transmission/reflection calculation 
+++++++++++++++++++++++++++++++++++

* :func:`.transmit4x4` to work with the computed system  matrix
* :func:`.transfer4x4` or :func:`.transfer` for higher level interface 

Intensity and Ez Hz field
+++++++++++++++++++++++++

* :func:`.poynting` the z component of the Poynting vector.
* :func:`.intensity` the absolute value of the Poytning vector.
* :func:`.EHz` for calculation of the z component of the E and H fields.

2x2 method
----------
    
todo..

"""

from __future__ import absolute_import, print_function, division

import numpy as np

from dtmm.conf import NCDTYPE,NFDTYPE, CDTYPE, FDTYPE, NUMBA_TARGET, \
                        NUMBA_PARALLEL, NUMBA_CACHE, NUMBA_FASTMATH, DTMMConfig
from dtmm.rotation import  _calc_rotations_uniaxial, _calc_rotations, _rotate_diagonal_tensor
from dtmm.linalg import _dotr2m, dotmdm, dotmm, inv, dotmv, _dotr2v
from dtmm.data import refind2eps
from dtmm.rotation import rotation_vector2
from dtmm.print_tools import print_progress

import numba as nb
from numba import prange
import time

if NUMBA_PARALLEL == False:
    prange = range

sqrt = np.sqrt

# available mode parameters.
_mode_int = {"r" : +1, "t" : +1, 1 : 1, -1 : -1}
_mode_int_none = {"b" : None, None : None}
_mode_int_none.update(_mode_int)

def _mode_to_int_or_none(mode):
    try:
        return _mode_int_none[mode]
    except KeyError:
        raise ValueError("Invalid propagation mode '{}'.".format(mode))

def _mode_to_int(mode):
    try :
        return _mode_int[mode]
    except KeyError:
        raise ValueError("Invalid propagation mode '{}'.".format(mode))


@nb.njit([(NFDTYPE,NCDTYPE[:],NCDTYPE[:,:])])                                                                
def _auxiliary_matrix(beta,eps,Lm):
    """Computes all elements of the auxiliary matrix of shape 4x4."""
    eps2m = 1./eps[2]
    eps4eps2m = eps[4]*eps2m
    eps5eps2m = eps[5]*eps2m
    
    Lm[0,0] = (-beta*eps4eps2m)
    Lm[0,1] = 1.-beta*beta*eps2m
    Lm[0,2] = (-beta*eps5eps2m)
    Lm[0,3] = 0.
    Lm[1,0] = eps[0]- eps[4]*eps4eps2m
    Lm[1,1] = Lm[0,0]
    Lm[1,2] = eps[3]- eps[5]*eps4eps2m
    Lm[1,3] = 0.
    Lm[2,0] = 0.
    Lm[2,1] = 0.
    Lm[2,2] = 0.
    Lm[2,3] = -1. 
    Lm[3,0] = (-1.0*Lm[1,2])
    Lm[3,1] = (-1.0*Lm[0,2])
    Lm[3,2] = beta * beta + eps[5]*eps5eps2m - eps[1]  
    Lm[3,3] = 0.  

@nb.njit([(NFDTYPE,NCDTYPE[:],NCDTYPE[:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _alphaf_iso(beta,eps0,alpha,F):
    """computes eigenvalue alpha and eigenvector field matrix of isotropic material"""
    #n = eps0[0]**0.5
    aout = sqrt(eps0[0]-beta**2)
    if aout != 0.:
        gpout = eps0[0]/aout
        gsout = -aout
        alpha[0] = aout
        alpha[1] = -aout
        alpha[2] = aout
        alpha[3] = -aout 
        F[0,0] = 0.5 
        F[0,1] = 0.5
        F[0,2] = 0.
        F[0,3] = 0.
        F[1,0] = 0.5 * gpout 
        F[1,1] = -0.5 * gpout 
        F[1,2] = 0.
        F[1,3] = 0.
        F[2,0] = 0.
        F[2,1] = 0.
        F[2,2] = 0.5 
        F[2,3] = 0.5
        F[3,0] = 0.
        F[3,1] = 0.
        F[3,2] = 0.5 * gsout 
        F[3,3] = -0.5 * gsout 
    else:
        
        F[...]=0.
        alpha[...] = 0.

@nb.njit([(NFDTYPE,NCDTYPE[:],NFDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:])], cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _alphaf_uniaxial(beta,eps0,R,alpha,F): 
    """computes eigenvalue alpha and eigenvector field matrix of uniaxial material"""
    #uniaxial case
    ct = R[2,2]
    st = -R[2,0] 
    st2 = st * st
    ct2 = ct * ct
    
    sf = -R[0,1]
    cf = R[1,1]

    eps11 = eps0[0]

    
    delta = eps0[2] -  eps11
    if beta == 0.: #same as calculation for beta !=0, except faster... no multiplying with zeros
        ev02 =  eps11 
        evs = sqrt(ev02)
        u = eps11 + delta * ct2
        w = eps11 * (ev02 + delta)
        sq = sqrt(u*w)/u
        evpp = sq
        evpm = -sq
        
    else: #can also be used for beta=0... just slower
        ev02 =  eps11 - beta * beta
        evs = sqrt(ev02)
        
        u = eps11 + delta * ct2
        gama = beta * cf
        v = gama * delta * 2 * st * ct
        w = gama * gama * (delta * st2)- eps11 * (ev02 + delta)
        
        sq = sqrt(v*v-4*u*w)/2/u
        v = v/2/u
        
        evpp = -v + sq
        evpm = -v - sq
        

    alpha[0] = evpp
    alpha[1] = evpm
    alpha[2] = evs
    alpha[3] = -evs    


    if beta == 0.:

        eps11sf = eps11 * sf
        evssf = evs*sf
        evscf = evs*cf
        eps11cf = eps11*cf
        
        F[0,2] = evssf
        F[1,2] = eps11sf
        F[2,2] = -evscf
        F[3,2] = eps11cf
        
        F[0,3] = -evssf
        F[1,3] = eps11sf 
        F[2,3] = evscf 
        F[3,3] = eps11cf
        
        F[0,0] = cf
        F[1,0] = evpp *cf
        F[2,0] = sf
        F[3,0] = -evpp *sf 
        
        F[0,1] = cf
        F[1,1] = evpm *cf
        F[2,1] = sf
        F[3,1] = -evpm *sf    
        
    else:
        sfst = (R[1,2])
        cfst = (R[0,2])                   
                                    
        ctbeta = ct * beta
        ctbetaeps11 = ctbeta / eps11
        eps11sfst = eps11 * sfst
        evssfst = evs*sfst
        evscfst = evs*cfst
        evsctbeta = evs*ctbeta
        ev02cfst = ev02*cfst
        ev02cfsteps11 = ev02cfst/eps11
      
        F[0,2] = -evssfst 
        F[1,2] = -eps11sfst
        F[2,2] = evscfst - ctbeta
        F[3,2] = evsctbeta - ev02cfst
  
        F[0,3] = -evssfst
        F[1,3] = eps11sfst
        F[2,3] = evscfst + ctbeta
        F[3,3] = ev02cfst + evsctbeta
        
        F[0,0] = -evpp*ctbetaeps11 + ev02cfsteps11
        F[1,0] = evpp *cfst - ctbeta
        F[2,0] = sfst
        F[3,0] = -evpp *sfst
        
        F[0,1] = -evpm*ctbetaeps11 + ev02cfsteps11
        F[1,1] = evpm *cfst - ctbeta
        F[2,1] = sfst
        F[3,1] = -evpm *sfst 
        
    #normalize base vectors
    for j in range(4):
        tmp = 0.
        for i in range(4):
            tmp += F[i,j].real * F[i,j].real + F[i,j].imag * F[i,j].imag
        
        tmp = tmp ** 0.5
        F[0,j] = F[0,j]/tmp 
        F[1,j] = F[1,j]/tmp 
        F[2,j] = F[2,j]/tmp 
        F[3,j] = F[3,j]/tmp 
        
        

@nb.njit([NFDTYPE(NCDTYPE[:])], cache = NUMBA_CACHE)
def _poynting(field):
    """Computes poynting vector from the field vector"""
    tmp1 = (field[0].real * field[1].real + field[0].imag * field[1].imag)
    tmp2 = (field[2].real * field[3].real + field[2].imag * field[3].imag)
    return tmp1-tmp2 

@nb.njit([(NCDTYPE[:],NCDTYPE[:,:],NCDTYPE[:],NCDTYPE[:,:])], cache = NUMBA_CACHE)
def _copy_sorted(alpha,fmat, out_alpha, out_fmat):
    """Eigen modes sorting based on the computed poynting vector direction"""
    i = 0
    j = 1
    
    # i,j,k,l = 0,1,2,3
    
    # p0 = _poynting(fmat[:,0])    
    # p1 = _poynting(fmat[:,1]) 
    # p2 = _poynting(fmat[:,2]) 
    # p3 = _poynting(fmat[:,3]) 
    
    # if p1 < p0:
    #     p0, p1 = p1, p0
    #     i, j = j, i
    # if p2 < p1:
    #     p1, p2 = p2, p1
    #     j, k = k, j
    # if p1 < p0:
    #     p0, p1 = p1, p0
    #     i, j = j, i
    # if p3 < p2:
    #     p3, p2 = p2, p3
    #     k,l = l, k
    # if p2 < p1:
    #     p1, p2 = p2, p1
    #     j, k = k, j        
    # if p1 < p0:
    #     p0, p1 = p1, p0
    #     i, j = j, i

    # out_alpha[i] = alpha[i]  
    # out_alpha[j] = alpha[j] 
    # out_alpha[k] = alpha[k] 
    # out_alpha[l] = alpha[l] 
    
    # out_fmat[:,i] = fmat[:,i]
    # out_fmat[:,j] = fmat[:,j]
    # out_fmat[:,k] = fmat[:,k]
    # out_fmat[:,l] = fmat[:,l]

    

    ok = True
    
 

    for k in range(4):
        p = _poynting(fmat[:,k])

        if p >= 0.:
            if i >=4:
                ok = False
            if ok:
                out_alpha[i] = alpha[k]
                out_fmat[:,i] = fmat[:,k]
            i = i + 2
        else:
            if j >=4:
                ok = False
            if ok : 
                out_alpha[j] = alpha[k]
                out_fmat[:,j] = fmat[:,k]
            j = j + 2
    if ok == False:
        print("Could not sort eigenvectors! Proceed with caution!")
        for i in range(4):
            #indicate that something went wrong, and that sorting was unsucesful
            out_alpha[i] = np.nan
            out_fmat[:,i] = 0
            
            
@nb.guvectorize([(NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NCDTYPE[:],NFDTYPE[:],NCDTYPE[:],NCDTYPE[:],NCDTYPE[:,:])],
                 "(),(),(m),(l),(k),(n)->(n),(n,n)", target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _alphaf_vec(beta,phi,rv,epsv,epsa,dummy,alpha,F):
    """eigenvalue solver. Depending on the material parameter
    we choose isotropic, uniaxial  or a biaxial solver.
    
    Becaue the auxiliary matrix is written in the rotated frame (in the plane of incidence with phi = 0)
    We need to rotate the computed vectors using _dotr2m 
    """
    #F is a 4x4 matrix... we can use 3x3 part for Rotation matrix and F[3] for eps  temporary data
    
    #isotropic case
    if (epsv[0] == epsv[1] and epsv[1]==epsv[2]):
        eps = F[3] 
        _alphaf_iso(beta[0],epsv,alpha,F)
        _dotr2m(rv,F,F)
    #uniaxial
    elif (epsv[0] == epsv[1]):
        R = F.real
        eps = F[3] 
        _calc_rotations_uniaxial(phi[0],epsa,R) #store rotation matrix in Fi.real[0:3,0:3]
        _alphaf_uniaxial(beta[0],epsv,R,alpha,F)
        _dotr2m(rv,F,F)
    else:#biaxial case
        R = F.real 
        eps = F.ravel() #reuse F memory (eps is length 6 1D array)
        _calc_rotations(phi[0],epsa,R) #store rotation matrix in Fi.real[0:3,0:3]
        _rotate_diagonal_tensor(R,epsv,eps)
        _auxiliary_matrix(beta[0],eps,F) #calculate Lm matrix and put it to F
        alpha0,F0 = np.linalg.eig(F)
        _copy_sorted(alpha0,F0,alpha,F)#copy data and sort it
        _dotr2m(rv,F,F)
        
#dummy arrays for gufuncs    
_dummy_array = np.empty((4,),CDTYPE)
_dummy_array2 = np.empty((9,),CDTYPE)
_dummy_EH = np.empty((2,),CDTYPE)
    

def _alphaf(beta,phi,epsv,epsa,out = None):
    rv = rotation_vector2(phi) 
    return _alphaf_vec(beta,phi,rv,epsv,epsa,_dummy_array, out = out)

def _default_beta_phi(beta, phi):
    """Checks the validity of beta, phi arguments and sets default values if needed"""
    beta = np.asarray(beta, FDTYPE) if beta is not None else np.asarray(0., FDTYPE)
    phi = np.asarray(phi, FDTYPE) if phi is not None else np.asarray(0., FDTYPE)
    return beta, phi

def _default_epsv_epsa(epsv, epsa):
    """Checks the validity of epsv, epsa arguments and sets default values if needed"""
    epsv = np.asarray(epsv, CDTYPE) if epsv is not None else np.asarray((1.,1.,1.), CDTYPE)
    epsa = np.asarray(epsa, FDTYPE) if epsa is not None else np.asarray((0.,0.,0.), FDTYPE)
    assert epsv.shape[-1] >= 3
    assert epsa.shape[-1] >= 3
    return epsv, epsa

def _as_field_vec(fvec):
    """converts input to valid field vector"""
    fvec = np.asarray(fvec, dtype = CDTYPE)
    assert fvec.shape[-1] == 4
    return fvec

def alphaf(beta = None, phi = None, epsv = None, epsa = None, out = None):
    """Computes alpha and field arrays (eigen values and eigen vectors arrays).
    
    Broadcasting rules apply.
    
    Parameters
    ----------
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    out : (ndarray,ndarray), optional
        Output arrays.
       
    Returns
    -------
    alpha, fieldmat: (ndarray, ndarray)
        Eigen values and eigen vectors arrays. 
    """
    beta, phi = _default_beta_phi(beta,phi)
    epsv, epsa = _default_epsv_epsa(epsv, epsa)
    rv = rotation_vector2(phi) 
    if out is None:
        return _alphaf_vec(beta,phi,rv,epsv,epsa,_dummy_array)
    else:
        return _alphaf_vec(beta,phi,rv,epsv,epsa,_dummy_array, out = out)

def alphaffi(beta=None,phi=None,epsv=None,epsa=None,out = None):
    """Computes alpha and field arrays (eigen values and eigen vectors arrays)
    and inverse of the field array. See also :func:`alphaf` 
    
    Broadcasting rules apply.
    
    Parameters
    ----------
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    out : (ndarray,ndarray,ndarray), optional
        Output arrays.
       
    Returns
    -------
    alpha, field, ifield  : (ndarray, ndarray, ndarray)
        Eigen values and eigen vectors arrays and its inverse
     
    Examples    
    --------
    
    This is equivalent to
    
    >>> alpha,field = alphaf(0,0, [2,2,2], [0.,0.,0.])
    >>> ifield = inv(field)
    """
    if out is not None:
        a,f,fi = out
        alphaf(beta,phi,epsv,epsa, out = (a,f))
        inv(f,fi)
    else:
        a,f = alphaf(beta,phi,epsv,epsa)
        fi = inv(f)
    return a,f,fi


 
def alphaE(beta,phi,epsv,epsa, mode = +1, out = None):
    """Computes E-field eigenvalue and eigenvector matrix for the 2x2 formulation.
    
    Parameters
    ----------
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : (ndarray,ndarray), optional
        Output arrays where results are written.
        
    Returns
    -------
    alpha, field : (ndarray, ndarray)
        Eigen values and eigen vectors arrays.
    
    """
    mode = _mode_to_int(mode)
    alpha,f = alphaf(beta,phi,epsv,epsa)
    e = E_mat(f,mode = mode, copy = False)
    if mode == 1:
        alpha = alpha[...,::2]
    else:
        alpha = alpha[...,1::2]
    if out is not None:
        out[0][...] = alpha
        out[1][...] = e
    else:
        out = alpha.copy(), e.copy()
    return out

def alphaEEi(beta,phi,epsv,epsa, mode = +1, out = None):
    """Computes E-field eigenvalue and eigenvector matrix and inverse of the 
    eigenvector array for the 2x2 formulation. See also :func:`alphaE` 
    
    Parameters
    ----------
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : (ndarray,ndarray,ndarray), optional
        Output arrays where results are written.

    Returns
    -------
    alpha, field, ifield : (ndarray, ndarray, ndarray)
        Eigen values and eigen vectors arrays and its inverse
    """
    mode = _mode_to_int(mode)
    if out is None:
        alpha,E = alphaE(beta,phi,epsv,epsa, mode = mode)
        return alpha, E, inv(E)
    else:
        alpha, E, Ei = out
        alpha, E = alphaE(beta,phi,epsv,epsa, mode = mode, out = (alpha, E))
        Ei = inv(E,out = Ei)
        return alpha, E, Ei 

from dtmm.conf import _numba_0_39_or_greater

if _numba_0_39_or_greater:
    @nb.vectorize([NCDTYPE(NCDTYPE,NFDTYPE)],
        target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)       
    def _phase_mat_vec(alpha,kd):
        return np.exp(NCDTYPE(1j)*kd*alpha)

    def _phasem(alpha,kd,out = None):
        kd = np.asarray(kd,FDTYPE)[...,None]
        out = _phase_mat_vec(alpha,kd,out)
        #if out.shape[-1] == 4:
        #    out[...,1::2]=0.
        return out
else:
    @nb.guvectorize([(NCDTYPE[:],NFDTYPE[:], NCDTYPE[:])],
                    "(n),()->(n)", target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)       
    def _phase_mat_vec(alpha,kd,out):
        for i in range(alpha.shape[0]):
            out[i] = np.exp(NCDTYPE(1j)*kd[0]*(alpha[i]))
                    
    _phasem = _phase_mat_vec

#to make autoapi happy...    
def phasem(*args,**kwargs):
    return _phasem(*args,**kwargs)
    
def phase_mat(alpha, kd, mode = None,  out = None):
    """Computes a 4x4 or 2x2 diagonal matrix from eigenvalue matrix alpha 
    and wavenumber. 
    
    The output is a diagonal, that is, a vector of length 2 or 4, depending on
    the input alpha array.
    
    Parameters
    ----------
    alpha : array
        The eigenvalue alpha array of shape (...,4) or (...,2).
    kd : float
        The kd phase value (layer thickness times wavenumber in vacuum).
    mode : int, optional
        If specified, converts the phase matrix to 2x2, taking either forward 
        propagating mode (+1), or negative propagating mode (-1).
    out : ndarray, optional
        Output array where results are written.
        
    Returns
    -------
    diag : array
        Phase diagonal matrix of shape (...,4) or (...,2).
    """
    mode = _mode_to_int_or_none(mode)
    alpha = np.asarray(alpha)
    
    if mode is not None:
        if alpha.shape[-1] != 4:
            raise ValueError("alpha array must be a 4-vector if mode is set.")
    
    kd = np.asarray(kd, dtype = FDTYPE)
    if out is None:
        if mode is None:
            b = np.broadcast(alpha,kd[...,None])
        else:
            b = np.broadcast(alpha[...,::2],kd[...,None])
        out = np.empty(b.shape, dtype = CDTYPE)
        
    if mode == +1:
        phasem(alpha[...,::2],kd, out = out)
    elif mode == -1:
        phasem(alpha[...,1::2],kd,out = out)
    elif mode is None:
        out = phasem(alpha,kd, out = out) 
    else:
        raise ValueError("Unknown propagation mode.")
    return out 

def iphase_mat(alpha, kd, cfact = 0.1, mode = +1,  out = None):
    """Computes incoherent 4x4 phase matrix from eigenvalue matrix alpha and wavenumber"""
    mode = _mode_to_int(mode)
    if mode == +1:
        np.add(alpha[...,1::2],alpha[...,1::2].real*2*1j*cfact, out = alpha[...,1::2])
    else:
        np.add(alpha[...,::2],-alpha[...,::2].real*2*1j*cfact, out = alpha[...,::2])
    return phase_mat(alpha, kd, mode = None, out = out)


@nb.guvectorize([(NCDTYPE[:], NFDTYPE[:])],
                    "(n)->()", target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)       
def poynting(fvec, out):
    """Calculates a z-component of the poynting vector from the field vector
    
    Parameters
    ----------
    fvec : (...,4,4) array
        Field matrix array.
    out : ndarray, optional
        Output array where results are written.
        
    Results
    -------
    poynting : array
        The z component of the poynting vector.
    """
    assert fvec.shape[0] == 4
    out[0] = _poynting(fvec)

def fmat2poynting(fmat, out = None):
    """Calculates poynting vectors (z component) from the field matrix.
    
    Parameters
    ----------
    fmat : (...,4,4) array
        Field matrix array.
    out : ndarray, optional
        Output array where results are written.
        
    Returns
    -------
    vec : (...,4) array
        Fmat's columns poynting vector z component.
    """
    axes = list(range(fmat.ndim))
    n = axes.pop(-2)
    axes.append(n)
    fmat = fmat.transpose(*axes)
    return poynting(fmat, out = out)
    
@nb.guvectorize([(NCDTYPE[:,:], NCDTYPE[:,:])],
                    "(n,n)->(n,n)", target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)       
def normalize_f(fmat, out):
    """Normalizes columns of field matrix so that fmat2poytning of the resulted
    matrix returns ones
    
    Parameters
    ----------
    fmat : (...,4,4) array
        Field matrix array.
    out : ndarray, optional
        Output array where results are written.
    """
    assert fmat.shape[0] == 4 and fmat.shape[1] == 4 
    for i in range(4):
        tmp1 = (fmat[0,i].real * fmat[1,i].real + fmat[0,i].imag * fmat[1,i].imag)
        tmp2 = (fmat[2,i].real * fmat[3,i].real + fmat[2,i].imag * fmat[3,i].imag)
        n = np.abs(tmp1-tmp2)**0.5 
        if n == 0.:
            n = 0.
        else:
            n = 1./n
        out[0,i] = fmat[0,i] * n 
        out[1,i] = fmat[1,i] * n
        out[2,i] = fmat[2,i] * n
        out[3,i] = fmat[3,i] * n


def intensity(fvec,out = None):
    """Calculates absolute value of the z-component of the poynting vector
    
    Parameters
    ----------
    fvec : (...,4) array
        Field vector array.
    out : ndarray, optional
        Output array where results are written.    
    """
    fvec = _as_field_vec(fvec)
    p = poynting(fvec)
    return np.abs(p)


def projection_mat(fmat, fmati = None, mode = +1, out = None):
    """Calculates projection matrix from the given field matrix. By multiplying
    the field with this matrix you obtain only the forward (mode = +1) or
    backward (mode = -1) propagating field,
    
    Parameters
    ----------

    fmat : (...,4,4) array
        Field matrix array.
    fmati : (...,4,4)
        The inverse of the field matrix. If not provided it is computed from `fmat`.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.   
    
    """
    mode = _mode_to_int(mode)
    if fmati is None:
        fmati = inv(fmat)
    diag = np.zeros(fmat.shape[:-1],fmat.dtype)
    if mode == 1:
        diag[...,0::2] = 1.
    else:
        diag[...,1::2] = 1.   
    return dotmdm(fmat,diag,fmati, out)    


    
def EHz(fvec, beta = None, phi = None, epsv = None, epsa = None, out = None):
    """Constructs the z component of the electric and magnetic fields 
    
    Parameters
    ----------
    fvec : (...,4,4) array
        Field matrix array.
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    out : (ndarray,ndarray), optional
        Output arrays where results are written.
        
    Returns
    -------
    Ez,Hz : (ndarray,ndarray)
        Ez and Hz arrays of shape (...,4)
    """
    fvec = _as_field_vec(fvec)
    beta, phi = _default_beta_phi(beta, phi)
    epsv,epsa = _default_epsv_epsa(epsv, epsa)
    rv = rotation_vector2(-phi)
    
    return _EHz(fvec,beta,phi,rv, epsv,epsa,_dummy_EH,out)
    
@nb.guvectorize([(NCDTYPE[:],NFDTYPE[:],NFDTYPE[:],NFDTYPE[:],NCDTYPE[:],NFDTYPE[:],NCDTYPE[:],NCDTYPE[:])],
                 "(n),(),(),(m),(l),(k),(o)->(o)", target = NUMBA_TARGET, cache = NUMBA_CACHE, fastmath = NUMBA_FASTMATH)
def _EHz(fvec, beta,phi,rv,epsv,epsa,dummy,out):
    eps = np.empty(shape = (6,), dtype = epsv.dtype)
    R = np.empty(shape = (3,3), dtype = rv.dtype)
    frot = np.empty_like(fvec)
    _dotr2v(rv,fvec,frot)
    _calc_rotations_uniaxial(phi[0],epsa,R)
    _rotate_diagonal_tensor(R,epsv,eps)
    out[0] = - (eps[4]*frot[0] + eps[5]*frot[2] + beta[0] * frot[1]) / eps[2]
    out[1] = beta[0] * frot[3]



def T_mat(fmatin, fmatout, fmatini = None, fmatouti = None, mode = +1):
    """Computes amplitude interface transmittance matrix.
    
    Parameters
    ----------
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    fmatouti : (...,4,4) array, optional
        Inverse of the output field matrix array. If not provided, it is computed
        from `fmatout`.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    """
    mode = _mode_to_int(mode)
    if fmatini is None:
        fmatini = inv(fmatin)
    if fmatouti is None:
        fmatouti = inv(fmatout)
    Sf = dotmm(fmatini,fmatout)
    Sb = dotmm(fmatouti,fmatin)
    out = np.zeros_like(Sf)
    if mode == +1:
        out[...,::2,::2] = inv(Sf[...,::2,::2])
        out[...,1::2,1::2] = Sb[...,1::2,1::2]
        return out
    else:
        out[...,1::2,1::2] = inv(Sf[...,1::2,1::2])
        out[...,::2,::2] = (Sb[...,::2,::2])
        return out
   
def S_mat(fmatin, fmatout, fmatini = None, overwrite_fmatin = False, mode = +1):
    """Computes the S matrix.
    
    Parameters
    ----------
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    fmatouti : (...,4,4) array, optional
        Inverse of the output field matrix array. If not provided, it is computed
        from `fmatout`.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    """
    mode = _mode_to_int(mode)
    if overwrite_fmatin == True:
        out = fmatin
    else:
        out = None
    if fmatini is None:
        fmatini = inv(fmatin, out = out)
    S = dotmm(fmatini,fmatout, out = out)
    if mode == +1:
        return S[...,::2,::2],S[...,1::2,0::2]
    else:
        return S[...,1::2,1::2],S[...,0::2,1::2]

# def transmission_mat(fmatin, fmatout, fmatini = None, mode = +1,out = None):
#     """Computes the transmission matrix.
    
#     Parameters
#     ----------
#     fmatin : (...,4,4) array
#         Input field matrix array.
#     fmatout : (...,4,4) array
#         Output field matrix array.
#     fmatini : (...,4,4) array
#         Inverse of the input field matrix array.
#     fmatouti : (...,4,4) array, optional
#         Inverse of the output field matrix array. If not provided, it is computed
#         from `fmatout`.
#     mode : int
#         Either +1, for forward propagating mode, or -1 for negative propagating mode.
#     out : ndarray, optional
#         Output array where results are written.
#     """
#     mode = _mode_to_int(mode)
#     A,B = S_mat(fmatin, fmatout, fmatini = fmatini, mode = mode)
#     if mode == +1:
#         A1 = fmatin[...,::2,::2]
#         A2 = fmatout[...,::2,::2]
#     else:
#         A1 = fmatin[...,1::2,1::2]
#         A2 = fmatout[...,1::2,1::2]      
#     Ai = inv(A, out = out)
#     A1i = inv(A1)
#     return dotmm(dotmm(A2,Ai, out = Ai),A1i, out = Ai)
#
#def reflection_mat(fin, fout, fini = None, mode = +1,out = None):
#    A,B = S_mat(fin, fout, fini = fini, mode = mode)
#    if mode == +1:
#        A1p = fin[...,::2,::2]
#        A1m = fin[...,::2,1::2]
#    elif mode == -1:
#        A1p = fin[...,1::2,1::2]
#        A1m = fin[...,1::2,::2]  
#    else:
#        raise ValueError("Unknown propagation mode.")
#    Ai = inv(A, out = out)
#    A1pi = inv(A1p)
#    return dotmm(dotmm(dotmm(A1m,B,out = Ai),Ai, out = Ai),A1pi, out = Ai)

def tr_mat(fmatin, fmatout, fmatini = None, overwrite_fmatin = False, mode = +1, out = None):
    """Computes the 2x2 tr matrix.
    
    Parameters
    ----------
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    overwrite_fmatin : bool
        Specifies whether fmatin can be overwritten or not.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : (ndarray,ndarray), optional
        Output arrays where results are written.
    """    
    mode = _mode_to_int(mode)
    if overwrite_fmatin == True:
        er = E_mat(fmatin, mode = mode * (-1), copy = True)
    else:
        er = E_mat(fmatin, mode = mode * (-1), copy = False)
    et = E_mat(fmatout, mode = mode, copy = False)
    eti,eri = Etri_mat(fmatin, fmatout, fmatini = fmatini, overwrite_fmatin = overwrite_fmatin, mode = mode, out = out)
    return dotmm(et,eti, out = eti), dotmm(er,eri, out = eri)

def t_mat(fmatin, fmatout, fmatini = None, overwrite_fmatin = False, mode = +1, out = None):
    """Computes the 2x2 tr matrix.
    
    Parameters
    ----------
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    overwrite_fmatin : bool
        Specifies whether fmatin can be overwritten or not.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.
    """   
    eti = Eti_mat(fmatin, fmatout, fmatini = fmatini, overwrite_fmatin = overwrite_fmatin, mode = mode, out = out)
    et = E_mat(fmatout, mode = mode, copy = False)
    return dotmm(et,eti, out = eti)

def E_mat(fmat, mode = None, copy = True):
    """Computes the E field matrix.
    
    Parameters
    ----------
    fmat : (...,4,4) array
        Field matrix array.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.

    """ 
    mode = _mode_to_int_or_none(mode)
    if mode == +1:
        e = fmat[...,::2,::2]
    elif mode == -1:
        e = fmat[...,::2,1::2]
    else:
        ep = fmat[...,::2,::2]
        en = fmat[...,::2,1::2]
        out = np.zeros_like(fmat)
        out[...,::2,::2] = ep
        out[...,1::2,1::2] = en
        return out 
    return e.copy() if copy else e  

def Eti_mat(fmatin, fmatout, fmatini = None, overwrite_fmatin = False, mode = +1, out = None):
    """Computes the inverse of the E field matrix (no reflections).
    
    Parameters
    ----------
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    overwrite_fmatin : bool
        Specifies whether fmatin can be overwritten or not.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.
    """     
    A = E_mat(fmatin, mode = mode, copy = False) 
    #Ai = inv(A, out = out)
    Ai = inv(A)
    St,Sr = S_mat(fmatin, fmatout, fmatini = fmatini, overwrite_fmatin = overwrite_fmatin, mode = mode)
    Sti = inv(St, out = St)
    return dotmm(Sti,Ai, out = out)


def Etri_mat(fmatin, fmatout, fmatini = None, overwrite_fmatin = False, mode = +1, out = None):
    """Computes the inverse of the E field matrix (with reflections).
    
    Parameters
    ----------
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    overwrite_fmatin : bool
        Specifies whether fmatin can be overwritten or not.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.
    """       
    out1, out2 = out if out is not None else (None, None)
    A = E_mat(fmatin, mode = mode, copy = False)
    Ai = inv(A, out = out1)  
    St,Sr = S_mat(fmatin, fmatout, fmatini = fmatini, overwrite_fmatin = overwrite_fmatin, mode = mode)
    Sti = inv(St, out = St)
    ei = dotmm(Sti,Ai,out = Ai)
    return ei, dotmm(Sr,ei, out = out2)
    
def E2H_mat(fmat, mode = +1, out = None): 
    """Computes the H field matrix from the field matrix.
    
    Parameters
    ----------
    
    fmat : (...,4,4) array
        Field matrix array.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.
    """ 
    mode = _mode_to_int(mode)      
    if mode == +1:
        A = fmat[...,::2,::2]
        B = fmat[...,1::2,::2]
    else:
        A = fmat[...,::2,1::2]
        B = fmat[...,1::2,1::2]
    Ai = inv(A, out = out)
    return dotmm(B,Ai, out = Ai)  

def f_iso(beta = 0., phi = 0., n = 1.):
    """Returns field matrix for isotropic layer of a given refractive index
    and beta, phi parameters
    
    Parameters
    ----------
    
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    n : float
        Refractive index of the medium (1. by default).
    """
    epsv = refind2eps([n]*3)
    epsa = np.zeros(shape = (3,),dtype= FDTYPE)
    alpha, f = alphaf(beta,phi,epsv,epsa)    
    return f

def ffi_iso(beta=0.,phi = 0., n=1):
    """Returns field matrix and inverse of the field matrix for isotropic layer 
    of a given refractive index and beta, phi parameters
    
    Parameters
    ----------
    
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    n : float
        Refractive index of the medium (1. by default).
    """
    epsv = refind2eps([n]*3)
    epsa = np.zeros(shape = (3,),dtype= FDTYPE)
    alpha, f, fi = alphaffi(beta,phi,epsv,epsa)    
    return f,fi

def layer_mat(kd, epsv,epsa, beta = 0,phi = 0, cfact = 0.1, method = "4x4", fmatin = None, retfmat = False, out = None):
    """Computes characteristic matrix of a single layer M=F.P.Fi,
    
    Numpy broadcasting rules apply
    
    Parameters
    ----------
    kd : float
        The kd phase value (layer thickness times wavenumber in vacuum).
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    cfact : float, optional
        Coherence factor, only used in combination with `4x4_2` method.
    method : str
        One of 4x4 (4x4 berreman - trasnmittance + reflectance), 
        2x2 (2x2 jones - transmittance only), 
        4x4_1 (4x4, single reflections - transmittance only),
        2x2_1 (2x2, single reflections - transmittance only) 
        4x4_2 (4x4, partially coherent reflections - transmittance only) 
    fmatin : ndarray, optional
        Used in combination with 2x2_1 method. It specifies the field matrix of 
        the input media in order to compute fresnel reflections. If not provided 
        it reverts to 2x2 with no reflections.
        
    out : ndarray, optional
    
    Returns
    -------
    cmat : ndarray
        Characteristic matrix of the layer.
    """    
    if method in ("2x2","2x2_1"):
        alpha,fmat = alphaf(beta,phi,epsv,epsa)
        f = E_mat(fmat, mode = +1, copy = False)
        if fmatin is not None and method == "2x2_1":
            fi = Eti_mat(fmatin, fmat, mode = +1)
        else:
            fi = inv(f)
        pmat = phase_mat(alpha[...,::2],kd)

    elif method in ("4x4","4x4_1","4x4_r","4x4_2"):
        alpha,fmat,fi = alphaffi(beta,phi,epsv,epsa)
        fmat = normalize_f(fmat)
        fi = inv(fmat)
        f = fmat
        pmat0 = phase_mat(alpha,-kd)
        if method ==  "4x4_r":
            alpha1 = alpha.copy()
            np.add(alpha1[...,1::2],alpha1[...,1::2].real*2j*cfact, out = alpha1[...,1::2])
            np.add(alpha[...,::2],-alpha[...,::2].real*2j*cfact, out = alpha[...,::2])
            pmat1 = phase_mat(alpha1,-kd)
            pmat2 = phase_mat(alpha,-kd)
            
            pmat = np.zeros(shape = pmat0.shape[:-1] + (8,8), dtype = pmat0.dtype)
            pmat[...,0,0] = pmat1[...,0]
            pmat[...,1,1] = pmat1[...,1]
            pmat[...,2,2] = pmat1[...,2]
            pmat[...,3,3] = pmat1[...,3]
            pmat[...,4,0] = pmat1[...,0]
            pmat[...,6,2] = pmat1[...,2]
            
            #pmat[...,4,4] = pmat2[...,0]
            #pmat[...,5,5] = pmat2[...,1]
            #pmat[...,6,6] = 0#pmat2[...,2]
            #pmat[...,7,7] = pmat2[...,3]
            pmat[...,5,1] = pmat2[...,1]
            pmat[...,7,3] = pmat2[...,3]
            #pmat[...,5,1] = ((1 - np.abs(pmat1[...,1])**2)**0.5)*pmat2[...,1]
            #pmat[...,7,3] = ((1 - np.abs(pmat1[...,3])**2)**0.5)*pmat2[...,3]
            
            f1 = np.zeros_like(pmat)
            f1[...,0:4,0:4] = f
            f1[...,4:8,4:8] = f
            
            f1i = inv(f1)
            #f1i = np.zeros_like(pmat)
            #f1i[...,0:4,0:4] = fi
            #f1i[...,4:8,4:8] = fi  
            
            return dotmm(f1,dotmm(pmat,f1i))
            
        elif method == "4x4_2":
            np.add(alpha[...,1::2],alpha[...,1::2].real*2*1j*cfact, out = alpha[...,1::2])

        pmat = phase_mat(alpha,-kd)
        if method == "4x4_1":
            pmat[...,1::2] = 0.
    else:
        raise ValueError("Unknown method!")
        
    out = dotmdm(f,pmat,fi,out = out) 
    
    if retfmat == False:
        return out   
    else:
        return fmat, out 

def stack_mat(kd,epsv,epsa, beta = 0, phi = 0, cfact = 0.01, method = "4x4", out = None):
    """Computes a stack characteristic matrix M = M_1.M_2....M_n if method is
    4x4, 4x2(2x4) and a characteristic matrix M = M_n...M_2.M_1 if method is
    2x2.
    
    Note that this function calls :func:`layer_mat`, so numpy broadcasting 
    rules apply to kd[i], epsv[i], epsa[i], beta and phi. 
    
    Parameters
    ----------
    kd : array of floats
        A sequence of phase values (layer thickness times wavenumber in vacuum).
        len(kd) must match len(epsv) and len(epsa).
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    cfact : float
        Coherence factor, only used in combination with `4x4_r` and `4x4_2` methods.
    method : str
        One of 4x4 (4x4 berreman), 2x2 (2x2 jones), 
        4x4_1 (4x4, single reflections), 2x2_1 (2x2, single reflections) 
        4x4_r (4x4, incoherent to compute reflection) or 
        4x4_t (4x4, incoherent to compute transmission) 
    out : ndarray, optional
    
    Returns
    -------
    cmat : ndarray
        Characteristic matrix of the stack.
    """
    t0 = time.time()
    mat = None
    tmat = None
    fmat = None
    n = len(kd)

    verbose_level = DTMMConfig.verbose
    if verbose_level > 1:
        print ("Building stack matrix.")
    for pi,i in enumerate(range(n)):
        print_progress(pi,n,level = verbose_level) 
        if method == "2x2_1":
            fmat, mat = layer_mat(kd[i],epsv[i],epsa[i],beta = beta, phi = phi, cfact = cfact, method = method, fmatin = fmat, out = mat, retfmat = True)
        else:
            mat = layer_mat(kd[i],epsv[i],epsa[i],beta = beta, phi = phi, cfact = cfact, method = method, out = mat)

        if pi == 0:
            if out is None:
                out = mat.copy()
            else:
                out[...] = mat
        else:
            if tmat is not None:
                dotmm(tmat,mat,mat)
            if method.startswith("2x2"):
                dotmm(mat,out,out)
            else:
                dotmm(out,mat,out)
    print_progress(n,n,level = verbose_level) 
    t = time.time()-t0
    if verbose_level >1:
        print("     Done in {:.2f} seconds!".format(t))  
    return out 

m1 = np.array([[1.,0,0,0],
         [0,1,0,0],
         [0,0,1,0],
         [0,0,0,1],
         [0,0,0,0],
         [0,0,0,0],
         [0,0,0,0],
         [0,0,0,0]])

m0 = np.array([[1.,0,0,0,0,0,0,0],
               [0,0,0,0,0,1,0,0],
               [0,0,1,0,0,0,0,0],
               [0,0,0,0,0,0,0,1]])    
    
def system_mat(cmat = None,fmatin = None, fmatout = None, fmatini = None, out = None):
    """Computes a system matrix from a characteristic matrix Fin-1.C.Fout
    
    Parameters
    ----------
    cmat : (...,4,4) array
        Characteristic matrix.
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    out : ndarray, optional
        Output array where results are written.
    """
    if fmatini is None:
        if fmatin is None:
            fmatin = f_iso()
        fmatini = inv(fmatin)
    if fmatout is None:
        fmatout = fmatin
    if cmat is not None:
        if cmat.shape[-1] == 8:
            dotmm(fmatini,cmat[...,0:4,0:4],out = cmat[...,0:4,0:4])
            dotmm(cmat[...,0:4,0:4],fmatout,out = cmat[...,0:4,0:4])
            dotmm(fmatini,cmat[...,4:8,4:8],out = cmat[...,4:8,4:8])
            dotmm(cmat[...,4:8,4:8],fmatout,out = cmat[...,4:8,4:8])
            if out is None:
                out = np.empty_like(cmat[...,0:4,0:4])
            
            out[...] = dotmm(m0,dotmm(cmat,m1))
            return out

        else:
            out = dotmm(fmatini,cmat,out = out)
            return dotmm(out,fmatout,out = out)  
    else:
        return dotmm(fmatini,fmatout,out = out)

def reflection_mat(smat, out = None):
    """Computes a 4x4 reflection matrix.
    
    Parameters
    ----------
    smat : (...,4,4) array
        System matrix.
    out : ndarray, optional
        Output array where results are written.
    """
    m1 = np.zeros_like(smat)
    m2 = np.zeros_like(smat)
    #fill diagonals
    for i in range(smat.shape[-1]//2):
        m1[...,i*2+1,i*2+1] = 1.
        m2[...,i*2,i*2] = -1.
    m1[...,:,0::2] = -smat[...,:,0::2]
    m2[...,:,1::2] = smat[...,:,1::2]
    m1 = inv(m1)
    return dotmm(m1,m2, out = out)

def fvec2E(fvec, fmat = None, fmati = None, mode = +1, inplace = False):
    """Converts field vector to E vector. If inplace == True, also 
    makes input field forward or backward propagating. 
    
    Parameters
    ----------
    fvec : (...,4,4) array
        Field vector array.
    fmat : (...,4,4) array
        Field matrix array.
    fmati : (...,4,4) array, optional
        The inverse of the field matrix. If not provided it is computed from `fmat`.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    inplace : bool, optional
    """
    mode = _mode_to_int(mode)
    if inplace == True:
        out  = fvec
    else:
        out = None
    if fmat is None:
        fmat = f_iso()
    pmat = projection_mat(fmat, fmati = fmati, mode = mode)
    return dotmv(pmat,fvec, out = out)[...,::2]
    # if mode == +1:
    #     return dotmv(pmat,fvec, out = out)[...,::2]
    # else:
    #     return dotmv(pmat,fvec, out = out)[...,1::2]

    
def E2fvec(evec, fmat = None, mode = +1, out = None):
    """Converts E vector to field vector
    
    Parameters
    ----------
    evec : (...,2) array
        E field vector array.
    fmat : (...,4,4) array
        Field matrix array.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.
    """
    evec = np.asarray(evec)
    if fmat is None:
        fmat = f_iso()
    e2h = E2H_mat(fmat, mode = mode)
    hvec = dotmv(e2h, evec)
    if out is None:
        out = np.empty(shape = evec.shape[:-1] + (4,), dtype = hvec.dtype) #hvec is complex
    out[...,::2] = evec
    out[...,1::2] = hvec
    return out
    
# def transmit2x2(evec_in, cmat,  tmatin = None, tmatout = None, evec_out = None):
#     """Transmits E-field vector using 2x2 method.
    
#     This functions takes an E-field vector that describes the input field and
#     computes the output transmited field using the 2x2 characteristic matrix.
    
#     Parameters
#     ----------
# {evec_in}
#     cmat : (...,4,4) array
#         Characteristic matrix.
# {tmatin}
# {tmatout}
# {evec_out}
    
#     """
#     b = np.broadcast(evec_in[...,0][...,None,None],cmat)

#     if evec_out is not None:
#         evec_out[...] = 0 
#     else:   
#         evec_out = np.zeros(b.shape[:-2] + (2,), evec_in.dtype)
        
#     if tmatin is not None:
#         evec = dotmv(tmatin, evec_in, out = evec_out)
#     else:
#         evec = evec_in
#     eout = dotmv(cmat, evec, out = evec_out)
#     if tmatout is not None:
#         eout = dotmv(tmatout, eout, out = evec_out)
#     return evec_out

def transmit2x2(fvec_in, cmat, fmatout = None, tmatin = None, tmatout = None, fvec_out = None):
    """Transmits field vector using 2x2 method.
    
    This functions takes a field vector that describes the input field and
    computes the output transmited field using the 2x2 characteristic matrix.
    
    Parameters
    ----------
    fvec_in : (...,4) array
        Input field vector array. This function will update the input array  
        with the calculated reflected field.
    cmat : (...,4,4) array
        Characteristic matrix.
    fmatout : (...,4,4) array
        Output field matrix array.
    tmatin : (...,2,2) array
        The transmittance matrix from the input medium to the first layer.
    tmatout : (...,2,2) array, optional
        The transmittance matrix from the last layer to the output maedium.
    fvec_out : (...,4) array, optional
        The ouptut field vector array. This function will update the output array 
        with the calculated transmitted field.
    
    """
    b = np.broadcast(fvec_in[...,0][...,None,None],cmat)
    if fmatout is None:
        fmatout = f_iso(1,0,0) 
    if fvec_out is not None:
        fvec_out[...] = 0 
    else:   
        fvec_out = np.zeros(b.shape[:-2] + (4,), fvec_in.dtype)
        
    if tmatin is not None:
        evec = dotmv(tmatin, fvec_in[...,::2], out = fvec_out[...,::2])
    else:
        evec = fvec_in[...,::2]
    eout = dotmv(cmat, evec, out = fvec_out[...,::2])
    if tmatout is not None:
        eout = dotmv(tmatout, eout, out = fvec_out[...,::2])
    e2h = E2H_mat(fmatout, mode = +1)
    hout = dotmv(e2h, eout, out = fvec_out[...,1::2])
    return fvec_out


def transmit4x4(fvec_in, cmat = None, fmatin = None, fmatout = None, fmatini = None, fmatouti = None, fvec_out = None):
    """Transmits field vector using 4x4 method.
    
    This functions takes a field vector that describes the input field and
    computes the output transmited field and also updates the input field 
    with the reflected waves.
   
    Parameters
    ----------
    fvec_in : (...,4) array
        Input field vector array. This function will update the input array  
        with the calculated reflected field
    cmat : (...,4,4) array
        Characteristic matrix.
    fmatin : (...,4,4) array
        Input field matrix array.
    fmatout : (...,4,4) array
        Output field matrix array.
    fmatini : (...,4,4) array
        Inverse of the input field matrix array.
    fmatouti : (...,4,4) array, optional
        Inverse of the output field matrix array. If not provided, it is computed
        from `fmatout`.
    fvec_out : (...,4) array, optional
        The ouptut field vector array. This function will update the output array 
        with the calculated transmitted field.
    """
    b = np.broadcast(fvec_in[..., None],cmat[...,0:4,0:4], fmatin, fmatout)
    
    if fvec_in.shape != b.shape[:-1]:
        raise ValueError("Input field vector should have shape of {}".format(b.shape[:-1]))
    if fvec_out is not None and fvec_out.shape != b.shape[:-1]:
        raise ValueError("Output field vector should have shape of {}".format(b.shape[:-1]))
    if fmatini is None:
        if fmatin is None:
            fmatin = f_iso()
        fmatini = inv(fmatin)
    if fmatin is None:
        fmatin = inv(fmatini)
    if fmatouti is None:
        if fmatout is None:
            fmatout = fmatin
            fmatouti = fmatini
        else:
            fmatouti = inv(fmatout)
    if fmatout is None:
        fmatout = inv(fmatouti)
        
    smat = system_mat(cmat = cmat,fmatini = fmatini, fmatout = fmatout)
     
    avec = dotmv(fmatini,fvec_in)
    
    
    a = np.zeros(b.shape[:-1], avec.dtype)
    a[...,0::2] = avec[...,0::2]
    avec = a.copy()#so that it broadcasts

    if fvec_out is not None:
        bvec = dotmv(fmatouti,fvec_out)
        a[...,1::2] = bvec[...,1::2] 
    else:
        bvec = np.zeros_like(avec)

    r = reflection_mat(smat)
    out = dotmv(r,a, out = fvec_out)
    
    avec[...,1::2] = out[...,1::2]
    bvec[...,::2] = out[...,::2]
        
    dotmv(fmatin,avec,out = fvec_in)    
    return dotmv(fmatout,bvec,out = out)

transmit = transmit4x4

def transfer4x4(fvec_in, kd, epsv, epsa,  beta = 0., phi = 0., nin = 1., nout = 1., 
             method = "4x4", reflect_in = False, reflect_out = False, fvec_out = None):
    """tranfers 4x4 field
    
    Parameters
    ----------
    fvec_in : (...,4) array
        Input field vector array. This function will update the input array  
        with the calculated reflected field
    kd : array of floats
        A sequence of phase values (layer thickness times wavenumber in vacuum).
        len(kd) must match len(epsv) and len(epsa).
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    nin : float
        Input layer refractive index.
    nout : float
        Output layer refractive index.
    method : str
        Any of 4x4, 4x4_1, 4x4_2, 4x4_r.
    reflect_in : bool
        Defines how to treat reflections from the input media and the first layer.
        If specified it does an incoherent reflection from the first interface.
    reflect_out : bool
        Defines how to treat reflections from the last layer and the output media.
        If specified it does an incoherent reflection from the last interface.
    fvec_out : (...,4) array, optional
        The ouptut field vector array. This function will update the output array 
        with the calculated transmitted field.
    """
    

    
    if method not in ("4x4", "4x4_1","4x4_r","4x4_2"):
        raise ValueError("Unknown method '{}'!".format(method))
        
        
    fveci = fvec_in
    fvecf = fvec_out
    
    fmatin = f_iso(n = nin, beta  = beta, phi = phi)
    fmatout = f_iso(n = nout, beta  = beta, phi = phi)
        
#    if reflect_in == True:
#        
#        #make fresnel reflection of the input (forward propagating) field
#        fin = f_iso(n = nin, beta  = beta, phi = phi)
#        alpha,fmatin = alphaf(beta = beta, phi = phi, epsv = epsv[0], epsa = epsa[0])
#        
#        t = T_mat(fin = fin, fout = fmatin, mode = +1)
#        fveci = dotmv(t,fveci, out = fveci)
#        
##        tmati = t_mat(fin,fmatin, mode = +1)
##        evec0 = fvec2E(fveci, fin, mode = +1)
##        evec = dotmv(tmati,evec0)
##        fveci = E2fvec(evec,fmatin, mode = +1)
##        fvec_in = E2fvec(evec0,fin, mode = +1, out = fvec_in)
#    else:
#        fmatin = f_iso(n = nin, beta  = beta, phi = phi)
#       
#    if reflect_out == True:
#        #make fresnel reflection of the output (backward propagating) field
#        alpha,fmatout = alphaf(beta = beta, phi = phi, epsv = epsv[-1], epsa = epsa[-1])
#        fout = f_iso(n = nout, beta  = beta, phi = phi)
#        if fvecf is not None:
#            t = T_mat(fin = fout, fout = fmatout, mode = -1)
#            fvecf = dotmv(t,fvecf)            
#            
##            tmatf = t_mat(fmatout,fout, mode = -1)
##            evec0 = fvec2E(fvecf, fout, mode = -1)
##            evec = dotmv(tmatf,evec) 
##            fvecf = E2fvec(evec,fmatout, mode = -1)
##            fvec_out = E2fvec(evec0,fout, mode = -1, out = fvec_out)
#    else:
#        fmatout = f_iso(n = nout, beta  = beta, phi = phi)
    
    cmat = stack_mat(kd, epsv, epsa, method = method)
    fvecf = transmit4x4(fveci, cmat = cmat, fmatin = fmatin, fmatout = fmatout, fvec_out = fvecf)

#    if reflect_in == True:
#        #make fresnel reflection of the input (backward propagating) field
#        t = T_mat(fin = fmatin, fout = fin, mode = -1)
#        fveci = dotmv(t,fveci, out = fvec_in)        
#        
##        tmati = t_mat(fin,fmatin, mode = -1)
##        evec = fvec2E(fveci, mode = -1)
##        evec = dotmv(tmati,evec)
##        fveci = E2fvec(evec,fout, mode = -1) 
##        np.add(fvec_in, fveci, out = fvec_in)
#
#
#    if reflect_out == True:
#        #make fresnel reflection of the output (forward propagating) field
#        
#        t = T_mat(fin = fmatout, fout = fout, mode = 1)
#        fvecf = dotmv(t,fvecf, out = fvec_out)    
##        tmati = t_mat(fmatout,fout, mode = +1)
##        evec = fvec2E(fvecf, mode = +1)
##        evec = dotmv(tmati,evec)
##        fvecf = E2fvec(evec,fout, mode = +1) 
##        fvecf = np.add(fvec_out, fvecf, out = fvec_out)  

    return fvecf

def transfer2x2(evec, kd, epsv, epsa,  beta = 0., phi = 0., nin  = None, nout = None, 
             method = "2x2", out = None):
    if method not in ("2x2", "2x2_1"):
        raise ValueError("Unknown method!")
        
    if nin is not None:
       fin = f_iso(n = nin, beta  = beta, phi = phi)
       alpha,fout = alphaf(beta = beta, phi = phi, epsv = epsv[0], epsa = epsa[0])
       tmat = t_mat(fin,fout)
       evec = dotmv(tmat,evec)
    
    cmat = stack_mat(kd, epsv, epsa, method = method)
    evec = dotmv(cmat,evec, out = out)
    
    if nout is not None:
       fout = f_iso(n = nout, beta  = beta, phi = phi)
       alpha,fin = alphaf(beta = beta, phi = phi, epsv = epsv[-1], epsa = epsa[-1])
       tmat = t_mat(fin,fout)
       dotmv(tmat,evec, out = evec)  
       
    return evec

def transfer(fvec_in, kd, epsv, epsa,  beta = 0., phi = 0., nin = 1., nout = 1., 
             method = "2x2", reflect_in = False, reflect_out = False, fvec_out = None):
    """Transfer input field vector through a layered material specified by the propagation
    constand k*d, eps tensor (epsv, epsa) and input and output isotropic media.
    
    Parameters
    ----------
    fvec_in : (...,4) array
        Input field vector array. This function will update the input array  
        with the calculated reflected field
    kd : array of floats
        A sequence of phase values (layer thickness times wavenumber in vacuum).
        len(kd) must match len(epsv) and len(epsa).
    epsv : (...,3) array, optional
        Dielectric tensor eigenvalues array (defaults to unity).
    epsa : (...,3) array, optional
        Euler rotation angles (psi, theta, phi) (defaults to (0.,0.,0.)).
    beta : float, optional
        The beta parameter of the field (defaults to 0.)
    phi : float, optional
        The phi parameter of the field (defaults to 0.)
    nin : float
        Input layer refractive index.
    nout : float
        Output layer refractive index.
    method : str
        Any of 4x4, 2x2, 2x2_1 or 4x4_1, 4x4_2, 4x4_r
    reflect_in : bool
        Defines how to treat reflections from the input media and the first layer.
        If specified it does an incoherent reflection from the first interface.
    reflect_out : bool
        Defines how to treat reflections from the last layer and the output media.
        If specified it does an incoherent reflection from the last interface.
    fvec_out : (...,4) array, optional
        The ouptut field vector array. This function will update the output array 
        with the calculated transmitted field.                
    """
    
    if method.startswith("2x2"):
        fin = f_iso(n = nin, beta  = beta, phi = phi)
        fout = f_iso(n = nout, beta  = beta, phi = phi)
        evec =  fvec2E(fvec_in, fmat = fin, mode = +1)
        nin = nin if reflect_in == True else None
        nout = nout if reflect_out == True else None
        evec = transfer2x2(evec, kd, epsv, epsa,  beta = beta, phi = phi, nin  = nin, nout = nout, 
             method = method)
        return E2fvec(evec,fout, mode = +1, out = fvec_out) 
    else:
        return transfer4x4(fvec_in, kd, epsv, epsa,  beta = beta, phi = phi, nin = nin, nout = nout, 
             method = method,  reflect_in = reflect_in, reflect_out = reflect_out, fvec_out = fvec_out)        
       
transfer1d = transfer
        
def avec(jones = (1,0), amplitude = 1., mode = +1, out = None):
    """Constructs amplitude vector.
    
    Numpy broadcasting rules apply for jones, and amplitude parameters
    
    Parameters
    ----------
    jones : jonesvec
        A jones vector, describing the polarization state of the field.
    amplitude : complex
        Amplitude of the field.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.
        
    Returns
    -------
    avec : ndarray
        Amplitude vector of shape (4,).
        
    Examples
    --------
    
    X polarized light with amplitude = 1
    >>> avec()
    array([1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j])
    
    X polarized light with amplitude 1 and y polarized light with amplitude 2.
    >>> b = avec(jones = ((1,0),(0,1)),amplitude = (1,2))
    >>> b[0]
    array([1.+0.j, 0.+0.j, 0.+0.j, 0.+0.j])
    >>> b[1]
    array([0.+0.j, 0.+0.j, 2.+0.j, 0.+0.j])
    """
    mode = _mode_to_int(mode)
    jones = np.asarray(jones)
    amplitude = np.asarray(amplitude)  
    c,s = jones[...,0], jones[...,1] 
    b = np.broadcast(c, amplitude)
    shape = b.shape + (4,)
    if out is None:
        out = np.empty(shape,CDTYPE)
    assert out.shape[-1] == 4
    if mode == +1:
        out[...,0] = c
        out[...,2] = s
        out[...,1] = 0.
        out[...,3] = 0.
    else:
        out[...,1] = c
        out[...,3] = s 
        out[...,0] = 0.
        out[...,2] = 0.

    out = np.multiply(out, amplitude[...,None] ,out = out)     
    return out

def fvec(fmat, jones = (1,0),  amplitude = 1., mode = +1, out = None):
    """Build field vector form a given polarization state, amplitude and mode.
    
    This function calls avec and then avec2fvec, see avec for details.
    
    Parameters
    ----------
    fmat : (...,4,4) array
        Field matrix array.
    jones : jonesvec
        A jones vector, describing the polarization state of the field.
    amplitude : complex
        Amplitude of the field.
    mode : int
        Either +1, for forward propagating mode, or -1 for negative propagating mode.
    out : ndarray, optional
        Output array where results are written.
    
    Returns
    -------
    fvec : ndarray
        Field vector of shape (...,4).
    
    Examples
    --------
    
    X polarized light traveling at beta = 0.4 and phi = 0.2 in medium with n = 1.5
    
    >>> fmat = f_iso(beta = 0.4, phi = 0.2, n = 1.5)
    >>> m = fvec(fmat, jones = jonesvec((1,0), phi = 0.2))
    
    This is equivalent to
    
    >>> a = avec(jones = jonesvec((1,0), phi = 0.2))
    >>> ma = avec2fvec(a,fmat)
    >>> np.allclose(ma,m)
    True
    """
    #a.shape != out.shape in general, so we are not using out argument here
    a = avec(jones, amplitude, mode)
    return avec2fvec(a, fmat, out = out)


def fvec2avec(fvec, fmat, normalize_fmat = True, out = None):
    """Converts field vector to amplitude vector
    
    Parameters
    ----------
    fvec : ndarray
        Input field vector
    fmat : ndarray
        Field matrix 
    normalize_fmat : bool, optional
        Setting this to false will not normalize the field matrix. In this case 
        user has to make sure that the normalization of the field matrix has 
        been performed prior to calling this function by calling normalize_f.
    out : ndarray, optional
        Output array
        
    Returns
    -------
    avec : ndarray
        Amplitude vector
    """
    if normalize_fmat == True:
        fmat = normalize_f(fmat)
    fmati = inv(fmat)
    return dotmv(fmati,fvec, out = out)

def avec2fvec(avec, fmat, normalize_fmat = True, out = None):
    """Converts amplitude vector to field vector
    
    Parameters
    ----------
    avec : ndarray
        Input amplitude vector
    fmat : ndarray
        Field matrix 
    normalize_fmat : bool, optional
        Setting this to false will not normalize the field matrix. In this case 
        user has to make sure that the normalization of the field matrix has 
        been performed prior to calling this function by calling normalize_f.
    out : ndarray, optional
        Output array
        
    Returns
    -------
    fvec : ndarray
        Field vector.
    """
    if normalize_fmat == True:
        fmat = normalize_f(fmat)
    return dotmv(fmat, avec, out = out)
    
__all__ = ["alphaf","alphaffi","phase_mat", "fvec", "avec", "fvec2avec",
           "avec2fvec","f_iso","ffi_iso","layer_mat","poynting","intensity",
           "transfer4x4","transmit4x4","transfer",
           "layer_mat","system_mat","stack_mat","EHz"]

if __name__ == "__main__":
    import doctest
    doctest.testmod()


import unittest
import numpy as np
import dtmm.wave as wave

class TestWave(unittest.TestCase):
    
    def setUp(self):
        self.shape = (64,64)
        self.ks = [1,2]
        self.beta, self.phi = wave.betaphi(self.shape,self.ks)

    def test_planewave_shapes(self):
        shapes = ((16,16), (15,15),(12,13),(7,6))
        k=1
        for shape in shapes:
            b,p = wave.betaphi(shape,k)
            for i in range(shape[0]):
                for j in range(shape[1]):
                    w1 = wave.planewave(shape,k,b[i,j],p[i,j])
                    w2 = wave.eigenwave(shape,i,j)
                    self.assertTrue(np.allclose(w1,w2))  

    def test_planewave_k(self):
        shape = (13,11)
        for k in (0.1,1.,10):
            b,p = wave.betaphi(shape,k)
            for i in range(shape[0]):
                for j in range(shape[1]):
                    w1 = wave.planewave(shape,k,b[i,j],p[i,j])
                    w2 = wave.eigenwave(shape,i,j)
                    self.assertTrue(np.allclose(w1,w2)) 

        
if __name__ == "__main__":
    unittest.main()


"""
The 2x2 layer propagation functions
"""
from __future__ import absolute_import, print_function, division

from dtmm.conf import BETAMAX
from dtmm.wave import eigenwave, betaphi
from dtmm.tmm import alphaf, E2H_mat, E_mat, Eti_mat, phase_mat, Etri_mat, tr_mat

from dtmm.linalg import dotmf, dotmdmf, inv, dotmd
from dtmm.diffract import diffract, E_tr_matrix
from dtmm.fft import fft2, ifft2
import numpy as np

from dtmm.matrix import corrected_E_diffraction_matrix,second_E_diffraction_matrix,first_E_diffraction_matrix
from dtmm.mode import fft_mask


def _transfer_ray_2x2_1(fft_field, wavenumbers, layer, effective_layer_in,effective_layer_out, dmat1, dmat2, beta = 0, phi=0,
                    nsteps = 1, mode = +1, reflection = True, betamax = BETAMAX, refl = None, bulk = None, out = None, tmpdata = None):
    _out = {} if tmpdata is None else tmpdata
    #fft_field = fft2(fft_field, out = out)
    shape = fft_field.shape[-2:]
    d_in, epsv_in,epsa_in = effective_layer_in     
    
    d_out, epsv_out,epsa_out = effective_layer_out    
    
    if reflection:
        tmat,rmat = E_tr_matrix(shape, wavenumbers, epsv_in = epsv_in, epsa_in = epsa_in,
                            epsv_out = epsv_out, epsa_out = epsa_out, mode = mode, betamax = betamax)
    
    d, epsv, epsa = layer
    alpha, fmat = alphaf(beta,phi, epsv, epsa, out = _out.get("alphaf"))
    
    e = E_mat(fmat, mode = mode, copy = False) #2x2 E-only view of fmat
    ei = inv(e, out = _out.get("ei"))
#    
    kd = wavenumbers * d   
    p = phase_mat(alpha,kd[...,None,None], mode = mode, out = _out.get("p"))
    
    if tmpdata is not None:
        _out["alphaf"] = alpha, fmat
        _out["ei"] = ei
        _out["p"] = p
    
    for j in range(nsteps):
        if j == 0 and reflection:
            #reflect only at the beginning
            if refl is not None:
                trans = refl.copy()
                refl = dotmf(rmat, fft_field, out = refl)
                fft_field = dotmf(tmat, fft_field, out = out)
                fft_field = np.add(fft_field,trans, out = fft_field)
                
                if mode == -1 and bulk is not None:
                    field = ifft2(fft_field)
                    e2h = E2H_mat(fmat, mode = mode)
                    bulk[...,::2,:,:] += field 
                    bulk[...,1::2,:,:] +=  dotmf(e2h, field, out = field)

                fft_field = dotmf(dmat1, fft_field, out = fft_field)
                out = fft_field
            else:
                fft_field = dotmf(tmat, fft_field, out = out)
                if dmat1 is not None:
                    fft_field = dotmf(dmat1, fft_field, out = fft_field)
                out = fft_field
        else:
            if dmat1 is not None:
                fft_field = dotmf(dmat1, fft_field, out = out)
            out = fft_field
        field = ifft2(fft_field, out = out)
        field = dotmdmf(e,p,ei,field, out = field)
        fft_field = fft2(field, out = field)
        if dmat2 is not None:
            fft_field = dotmf(dmat2, fft_field, out = fft_field)
    #return fft_field, refl  
    
    #out = ifft2(fft_field, out = out)
   
    if mode == +1 and bulk is not None:
        field = ifft2(fft_field)
        e2h = E2H_mat(fmat, mode = mode)
        bulk[...,1::2,:,:] +=  dotmf(e2h, field)
        bulk[...,::2,:,:] += field
    
    return fft_field, refl



def _transfer_ray_2x2_2(field, wavenumbers, in_layer, out_layer, dmat = None, beta = 0, phi=0,
                    nsteps = 1, mode = +1,  reflection = True, betamax = BETAMAX, refl = None, bulk = None, out = None, tmpdata = None):
    _out = {} if tmpdata is None else tmpdata
    if in_layer is not None:
        d, epsv,epsa = in_layer    
        alpha, fmat_in = alphaf(beta,phi, epsv, epsa, out = _out.get("afin"))
        if tmpdata is not None:
            _out["afin"] = alpha, fmat_in
    d, epsv,epsa = out_layer        
    alpha, fmat = alphaf(beta,phi, epsv, epsa, out = _out.get("afout"))    
    e = E_mat(fmat, mode = mode, copy = False) #2x2 E-only view of fmat
#    if refl is not None:
#        ein = E_mat(fmat_in, mode = mode * (-1)) #reverse direction
#
#    if reflection == 0:
#        kd = wavenumbers * d
#    else:   
    kd = wavenumbers * d /2  
    p = phase_mat(alpha,kd[...,None,None], mode = mode, out = _out.get("p"))

    ei0 = inv(e, out = _out.get("ei0"))
    ei = ei0

    if tmpdata is not None:
        _out["afout"] = alpha, fmat
        _out["ei0"] = ei
        _out["p"] = p
    
    for j in range(nsteps):
        #reflect only at the beginning
        if j == 0 and reflection != 0:
            #if we need to track reflections (multipass)
            if refl is not None:
                #ei,eri = Etri_mat(fmat_in, fmat, mode = mode, out = _out.get("eieri"))
                tmat,rmat = tr_mat(fmat_in, fmat, mode = mode, out = _out.get("eieri"))
                if tmpdata is not None:
                    #_out["eieri"] = ei,eri
                    _out["eieri"] = tmat, rmat
                trans = refl.copy()
 
                #refl = dotmf(rmat, field, out = refl)
                refl = dotmf(tmat, field, out = refl)
                refl = np.subtract(field, refl, out = refl)

                field = dotmf(tmat,field, out = out)
                field = np.add(field,trans, out = field)
                
                
                if mode == -1 and bulk is not None:
                    #tmp_field = dotmf(e,field)
                    tmp_field = field
                    e2h = E2H_mat(fmat, mode = mode)
                    bulk[...,::2,:,:] += tmp_field 
                    bulk[...,1::2,:,:] +=  dotmf(e2h, tmp_field)
                
                field = dotmf(ei,field, out = field)
                
                if d != 0.:                
                    field = dotmf(dotmd(e,p),field, out = field)
                    
                else:
                    field = dotmf(e,field, out = field)
                out = field   
                
#                rmat = dotmm(ein, eri, out = eri)
#                trans = refl.copy()
#                refl = dotmf(rmat, field, out = refl)     
#                ei0 = inv(e)
#                field = dotmf(ei,field, out = out)
#                field = dotmf(e,field) + trans
#                if d != 0.:
#                    field = dotmdmf(e,p,ei0,field, out = out)
                    
                ei = ei0
            else:
                ei = Eti_mat(fmat_in, fmat, mode = mode, out = _out.get("eti"))
                field = dotmdmf(e,p,ei,field, out = out)  
                out = field
                if tmpdata is not None:
                    _out["eti"] = ei
                ei = ei0
        else:
            #no need to compute if d == 0!.. identity
            if d != 0.:
                field = dotmdmf(e,p,ei,field, out = out) 
                out = field
        if dmat is not None:   
            field = diffract(field, dmat, out = out)
            out = field
        #no need to compute if d == 0!.. identity
        if d != 0.:
            field = dotmdmf(e,p,ei,field, out = out) 
            
    if mode == +1 and bulk is not None:
        e2h = E2H_mat(fmat, mode = mode)
        bulk[...,1::2,:,:] +=  dotmf(e2h, field)
        bulk[...,::2,:,:] += field
     

    return field, refl

        
def propagate_2x2_effective_1(field, wavenumbers, layer_in, layer_out, effective_layer_in, 
                            effective_layer_out, beta = 0, phi = 0,
                            nsteps = 1, diffraction = True, reflection = True, 
                            betamax = BETAMAX,  mode = +1,  tmpdata = None, split_diffraction = False,
                            refl = None, bulk = None, out = None):
    d_out, epsv_out,epsa_out = effective_layer_out    
    shape = field.shape[-2:]
    
    if diffraction <= 1:
        if diffraction == 1:
            dmat1 = first_E_diffraction_matrix(shape, wavenumbers, beta, phi,d_out/2, epsv = epsv_out, 
                                        epsa =  epsa_out, mode = mode, betamax = betamax) 
            dmat2 = second_E_diffraction_matrix(shape, wavenumbers, beta, phi,d_out/2, epsv = epsv_out, 
                                        epsa =  epsa_out, mode = mode, betamax = betamax) 
        else:
            dmat1, dmat2 = None,None
        return _transfer_ray_2x2_1(field, wavenumbers, layer_out, effective_layer_in, effective_layer_out,dmat1, dmat2,
                                beta = beta, phi = phi, nsteps =  nsteps,reflection = reflection,
                                betamax = betamax, mode = mode, refl = refl, bulk = bulk, out = out, tmpdata = tmpdata)            
    elif diffraction > 1:
        fout = np.zeros_like(field)
        reflpart = None
        fpart = None
        _out = None

        if refl is not None:
            _refl = np.zeros_like(refl)
        else:
            _refl = None

        try: 
            broadcast_shape = beta.shape
            beta = beta[...,0]
            phi = phi[...,0]
        except IndexError:
            broadcast_shape = ()
            
        windows, (betas, phis) = fft_mask(field.shape, wavenumbers, int(diffraction), 
                 betax_off = beta*np.cos(phi), betay_off = beta*np.sin(phi), betamax = betamax)    

        n = len(windows)
        betas = betas.reshape((n,) + broadcast_shape)
        phis = phis.reshape((n,) + broadcast_shape)
        
        if split_diffraction == False:

            dmats1 = first_E_diffraction_matrix(shape, wavenumbers, betas, phis,d_out/2, epsv = epsv_out, 
                                        epsa =  epsa_out, mode = mode, betamax = betamax) 
            dmats2 = second_E_diffraction_matrix(shape, wavenumbers, betas, phis,d_out/2, epsv = epsv_out, 
                                        epsa =  epsa_out, mode = mode, betamax = betamax) 
            
            idata = zip(windows, betas, phis, dmats1, dmats2)
        else:
            idata = zip(windows, betas, phis)
            
        for data  in idata:
            if split_diffraction == False:
                window, beta, phi, dmat1, dmat2 = data
            else:
                window, beta, phi = data
                dmat1 = first_E_diffraction_matrix(shape, wavenumbers, beta, phi,d_out/2, epsv = epsv_out, 
                                        epsa =  epsa_out, mode = mode, betamax = betamax) 
                dmat2 = second_E_diffraction_matrix(shape, wavenumbers, beta, phi,d_out/2, epsv = epsv_out, 
                                        epsa =  epsa_out, mode = mode, betamax = betamax) 
                        
            
            fpart = np.multiply(field, window, out = fpart)
            
            if refl is not None:
                reflpart = np.multiply(refl, window, out = reflpart)
            else:
                reflpart = None
            _out, __refl = _transfer_ray_2x2_1(fpart, wavenumbers,layer_out, 
                                                effective_layer_in, effective_layer_out,
                                                dmat1, dmat2,
                            beta = beta, phi = phi, 
                            nsteps =  nsteps,reflection = reflection,
                            betamax = betamax, mode = mode, bulk = bulk,
                            out = _out,  refl = reflpart, tmpdata = tmpdata)
             
            np.add(fout, _out, fout)
            if refl is not None and reflection != 0:
                np.add(_refl, __refl, out = _refl)


    if out is not None:
        out[...] = fout
    else:
        out = fout
    if refl is not None:
        refl[...] = _refl
    return out, refl   

def propagate_2x2_effective_2(field, wavenumbers, layer_in, layer_out, effective_layer_in, 
                            effective_layer_out, beta = 0, phi = 0,
                            nsteps = 1, diffraction = True, split_diffraction = False,
                            reflection = True, 
                            betamax = BETAMAX,  mode = +1, 
                            refl = None, bulk = None, out = None, tmpdata = None):
    
    shape = field.shape[-2:]
    d_eff, epsv_eff, epsa_eff = effective_layer_out
    

    
    if diffraction <= 1:
        if diffraction:
            dmat = corrected_E_diffraction_matrix(shape, wavenumbers, beta,phi, d = d_eff,
                                 epsv = epsv_eff, epsa = epsa_eff, mode = mode, betamax = betamax)
        else:
            dmat = None
        
        return _transfer_ray_2x2_2(field, wavenumbers, layer_in, layer_out, dmat = dmat,
                                beta = beta, phi = phi, nsteps =  nsteps,
                                reflection = reflection,
                                betamax = betamax, mode = mode, refl = refl, bulk = bulk, out = out, tmpdata = tmpdata)            
    else:
        fout = 0.
        fpart = None
        ffield = fft2(field)
        if refl is not None:
            frefl = fft2(refl)
            _refl = 0.
        else:
            _refl = None
        try: 
            broadcast_shape = beta.shape
            beta = beta[...,0]
            phi = phi[...,0]
        except IndexError:
            #beta and phi are scalar
            broadcast_shape = ()
            
        windows, (betas, phis) = fft_mask(field.shape, wavenumbers, int(diffraction), 
                 betax_off = beta*np.cos(phi), betay_off = beta*np.sin(phi), betamax = betamax)    

        n = len(windows)
        betas = betas.reshape((n,) + broadcast_shape)
        phis = phis.reshape((n,) + broadcast_shape)


        if split_diffraction == False:

            dmats = corrected_E_diffraction_matrix(shape, wavenumbers, betas,phis, d = d_eff,
                                     epsv = epsv_eff, epsa = epsa_eff, mode = mode, betamax = betamax)
            
            idata = zip(windows, betas, phis, dmats)
        else:
            idata = zip(windows, betas, phis)
            

        for data  in idata:
            if split_diffraction == False:
                window, beta, phi, dmat = data  
            else:
                window, beta, phi = data
                dmat = corrected_E_diffraction_matrix(shape, wavenumbers, beta,phi, d = d_eff,
                                     epsv = epsv_eff, epsa = epsa_eff, mode = mode, betamax = betamax)
                        
            fpart = np.multiply(ffield, window, out = fpart)
            fpart_re = ifft2(fpart, out = fpart)
            
            if refl is not None:
                reflpart = frefl* window
                reflpart_re = ifft2(reflpart)
            else:
                reflpart_re = None
            _out, __refl = _transfer_ray_2x2_2(fpart_re, wavenumbers, layer_in, layer_out, dmat = dmat,
                                beta = beta, phi = phi, nsteps =  nsteps,
                                betamax = betamax, reflection = reflection,
                                mode = mode, bulk = bulk,out = out,  refl = reflpart_re,tmpdata = tmpdata)                       
    
            fout += _out
            if refl is not None and reflection != 0:
                _refl += __refl

    if out is not None:
        out[...] = fout
    else:
        out = fout
    if refl is not None:
        refl[...] = _refl

    return out, refl
 
    
def propagate_2x2_full(field, wavenumbers, layer, input_layer = None, 
                    nsteps = 1,  mode = +1, reflection = True,
                    betamax = BETAMAX, refl = None, bulk = None, out = None):

    shape = field.shape[-2:]
   
    d, epsv, epsa = layer
    if input_layer is not None:
        d_in, epsv_in, epsa_in = input_layer

    kd = wavenumbers*d/nsteps
    
    if out is None:
        out = np.empty_like(field)
        
    ii,jj = np.meshgrid(range(shape[0]), range(shape[1]),copy = False, indexing = "ij") 
    
    for step in range(nsteps):
        for i in range(len(wavenumbers)):
            ffield = fft2(field[...,i,:,:,:])
            ofield = np.zeros_like(out[...,i,:,:,:])

            b,p = betaphi(shape,wavenumbers[i])
            mask = b < betamax
            
            amplitude = ffield[...,mask]
            
            betas = b[mask]
            phis = p[mask]
            iind = ii[mask]
            jind = jj[mask]
            
            if bulk is not None:
                obulk = bulk[...,i,:,:,:]
            
            if refl is not None:
                tampl = fft2(refl[...,i,:,:,:])[...,mask]
                orefl = refl[...,i,:,:,:]
                orefl[...] = 0.
              
            for bp in sorted(zip(range(len(betas)),betas,phis,iind,jind),key = lambda el : el[1], reverse = False):     
                #for j,bp in enumerate(zip(betas,phis,iind,jind)):     
                              
                j, beta, phi, ieig, jeig = bp

                out_af = alphaf(beta,phi,epsv,epsa)
                alpha,fmat_out = out_af 
                e = E_mat(fmat_out, mode = mode)
                ei0 = inv(e)
                ei = ei0                        
                pm = phase_mat(alpha,kd[i,None,None],mode = mode)
                w = eigenwave(amplitude.shape[:-1]+shape, ieig,jeig, amplitude = amplitude[...,j])
                if step == 0 and reflection != False:
                    alphain, fmat_in = alphaf(beta,phi,epsv_in,epsa_in)
                    if refl is not None:
                        ei,eri = Etri_mat(fmat_in, fmat_out, mode = mode)
                        ein =  E_mat(fmat_in, mode = -1*mode)
                        t = eigenwave(amplitude.shape[:-1]+shape, ieig,jeig, amplitude = tampl[...,j])
                        r = dotmf(eri, w)
                        r = dotmf(ein,r, out = r)
                        np.add(orefl,r,orefl)
                    
                        w = dotmf(ei, w, out = w)
                        t = dotmf(ei0,t, out = t)
                        w = np.add(t,w,out = w)
  
                    else:
                        ei = Eti_mat(fmat_in, fmat_out, mode = mode)
                        w = dotmf(ei, w, out = w)
                    w = dotmf(dotmd(e,pm),w, out = w)
                    np.add(ofield,w,ofield)                         
                    
                else:
                    w = dotmdmf(e,pm,ei,w, out = w)
                    np.add(ofield,w,ofield)

                if bulk is not None:
                    e2h = E2H_mat(fmat_out, mode = mode)
                    obulk[...,1::2,:,:] +=  dotmf(e2h, w)
                    obulk[...,::2,:,:] += w
                
            out[...,i,:,:,:] = ofield
                    
        field = out
    return out, refl

